"""PTF workflows for Wflow plugin."""

import math

import numpy as np


def kv_brakensiek(thetas, clay, sand):
    """
    Determine saturated hydraulic conductivity kv [mm/day].

    Based on:
      Brakensiek, D.L., Rawls, W.J.,and Stephenson, G.R.: Modifying scs hydrologic
      soil groups and curve numbers for range land soils, ASAE Paper no. PNR-84-203,
      St. Joseph, Michigan, USA, 1984.

    Parameters
    ----------
    thetas: float
        saturated water content [m3/m3].
    clay : float
        clay percentage [%].
    sand: float
        sand percentage [%].

    Returns
    -------
    kv : float
        saturated hydraulic conductivity [mm/day].

    """
    kv = (
        np.exp(
            19.52348 * thetas
            - 8.96847
            - 0.028212 * clay
            + (1.8107 * 10**-4) * sand**2
            - (9.4125 * 10**-3) * clay**2
            - 8.395215 * thetas**2
            + 0.077718 * sand * thetas
            - 0.00298 * sand**2 * thetas**2
            - 0.019492 * clay**2 * thetas**2
            + (1.73 * 10**-5) * sand**2 * clay
            + 0.02733 * clay**2 * thetas
            + 0.001434 * sand**2 * thetas
            - (3.5 * 10**-6) * clay**2 * sand
        )
        * (2.78 * 10**-6)
        * 1000
        * 3600
        * 24
    )

    return kv


def kv_cosby(sand, clay):
    """
    Determine saturated hydraulic conductivity kv [mm/day].

    based on:
      Cosby, B.J., Hornberger, G.M., Clapp, R.B., Ginn, T.R., 1984.
      A statistical exploration of the relationship of soil moisture characteristics to
      the physical properties of soils. Water Resour. Res. 20(6) 682-690.

    Parameters
    ----------
    sand: float
        sand percentage [%].
    clay : float
        clay percentage [%].

    Returns
    -------
    kv : float
        saturated hydraulic conductivity [mm/day].

    """
    kv = 60.96 * 10.0 ** (-0.6 + 0.0126 * sand - 0.0064 * clay) * 10.0

    return kv


def pore_size_index_brakensiek(sand, thetas, clay):
    """
    Determine Brooks-Corey pore size distribution index [-].

    Based on:
      Rawls,W. J., and Brakensiek, D. L.: Estimation of SoilWater Retention and
      Hydraulic Properties, In H. J. Morel-Seytoux (Ed.),
      Unsaturated flow in hydrologic modelling - Theory and practice, NATO ASI Series 9,
      275–300, Dordrecht, The Netherlands: Kluwer Academic Publishing, 1989.

    Parameters
    ----------
    sand: float
        sand percentage [%].
    thetas : float
        saturated water content [m3/m3].
    clay: float
        clay percentage [%].

    Returns
    -------
    poresizeindex : float
        pore size distribution index [-].

    """
    poresizeindex = np.exp(
        -0.7842831
        + 0.0177544 * sand
        - 1.062498 * thetas
        - (5.304 * 10**-5) * (sand**2)
        - 0.00273493 * (clay**2)
        + 1.11134946 * (thetas**2)
        - 0.03088295 * sand * thetas
        + (2.6587 * 10**-4) * (sand**2) * (thetas**2)
        - 0.00610522 * (clay**2) * (thetas**2)
        - (2.35 * 10**-6) * (sand**2) * clay
        + 0.00798746 * (clay**2) * thetas
        - 0.00674491 * (thetas**2) * clay
    )

    return poresizeindex


def thetas_toth(ph, bd, clay, silt):
    """
    Determine saturated water content [m3/m3].

    Based on:
      Tóth, B., Weynants, M., Nemes, A., Makó, A., Bilas, G., and Tóth, G.:
      New generation of hydraulic pedotransfer functions for Europe, Eur. J.
      Soil Sci., 66, 226–238. doi: 10.1111/ejss.121921211, 2015.

    Parameters
    ----------
    ph: float
        pH [-].
    bd : float
        bulk density [g /cm3].
    sand: float
        sand percentage [%].
    silt: float
        silt percentage [%].

    Returns
    -------
    thetas : float
        saturated water content [cm3/cm3].

    """
    thetas = (
        0.5653
        - 0.07918 * bd**2
        + 0.001671 * ph**2
        + 0.0005438 * clay
        + 0.001065 * silt
        + 0.06836
        - 0.00001382 * clay * ph**2
        - 0.00001270 * silt * clay
        - 0.0004784 * bd**2 * ph**2
        - 0.0002836 * silt * bd**2
        + 0.0004158 * clay * bd**2
        - 0.01686 * bd**2
        - 0.0003541 * silt
        - 0.0003152 * ph**2
    )

    return thetas


def thetar_rawls_brakensiek(sand, clay, thetas):
    """
    Determine residual water content [m3/m3].

    Based on:
      Rawls,W. J., and Brakensiek, D. L.: Estimation of SoilWater Retention and
      Hydraulic Properties, In H. J. Morel-Seytoux (Ed.),
      Unsaturated flow in hydrologic modelling - Theory and practice, NATO ASI Series 9,
      275–300, Dordrecht, The Netherlands: Kluwer Academic Publishing, 1989.

    Parameters
    ----------
    sand : float
        sand percentage [%].
    clay: float
        clay percentage [%].
    thetas: float
        saturated water content [m3/m3].

    Returns
    -------
    thetar : float
        residual water content [m3/m3].

    """
    thetar = (
        -0.0182482
        + 8.7269 * 10**-4 * sand
        + 0.00513488 * clay
        + 0.02939286 * thetas
        - 1.5395 * 10**-4 * clay**2
        - 1.0827 * 10**-3 * sand * thetas
        - 1.8233 * 10**-4 * clay**2 * thetas**2
        + 3.0703 * 10**-4 * clay**2 * thetas
        - 2.3584 * 10**-3 * thetas**2 * clay
    )

    return thetar


def soil_texture_usda(clay, silt):
    """
    Determine USDA soil texture.

    Parameters
    ----------
    clay: float
        sand percentage [%].
    silt: float
        silt percentage [%].

    Returns
    -------
    soil texture : int
        based on integer mapping following Ballabio et al. 2016 \
(https://doi.org/10.1016/j.geoderma.2015.07.006) for \
European topsoil physical properties.
        Value	NAME
        1	Clay
        2	Silty Clay
        3	Silty Clay-Loam
        4	Sandy Clay
        5	Sandy Clay-Loam
        6	Clay-Loam
        7	Silt
        8	Silt-Loam
        9	Loam
        10	Sand
        11	Loamy Sand
        12	Sandy Loam


    """
    sand = 100 - (clay + silt)

    soil_texture = np.where(
        np.logical_and(clay >= 40.0, sand >= 20.0, sand <= 45),
        1,  # clay
        np.where(
            np.logical_and(clay >= 27.0, sand >= 20.0, sand <= 45),
            6,  # clay loam
            np.where(
                np.logical_and(silt <= 40.0, sand <= 20.0),
                1,  # clay
                np.where(
                    np.logical_and(silt > 40.0, clay >= 40.0),
                    2,  # silty clay
                    np.where(
                        np.logical_and(clay >= 35.0, sand >= 45.0),
                        4,  # sandy clay
                        np.where(
                            np.logical_and(clay >= 27.0, sand < 20.0),
                            3,  # silty clay loam
                            np.where(
                                np.logical_and(clay <= 10.0, silt >= 80.0),
                                7,  # silt
                                np.where(
                                    (silt >= 50.0),
                                    8,  # silt loam
                                    np.where(
                                        np.logical_and(
                                            clay >= 7.0, sand <= 52.0, silt >= 28.0
                                        ),
                                        9,  # loam
                                        np.where(
                                            (clay >= 20.0),
                                            5,  # sandy clay loam
                                            np.where(
                                                (clay >= (sand - 70)),
                                                12,  # sandy loam
                                                np.where(
                                                    (clay >= (2 * sand - 170.0)),
                                                    11,  # loamy sand
                                                    np.where(
                                                        np.isnan(clay), np.nan, 10
                                                    ),  # sand
                                                ),
                                            ),
                                        ),
                                    ),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
        ),
    )

    return soil_texture


def ErosK_texture(clay, silt):
    """
    Determine mean detachability of the soil (Morgan et al., 1998).

    Based on USDA soil texture.

    Parameters
    ----------
    clay: float
        sand percentage [%].
    silt: float
        silt percentage [%].

    Returns
    -------
    ErosK : float
        mean detachability of the soil [g/J].

    """
    sand = 100 - (clay + silt)

    erosK = np.where(
        np.logical_and(clay >= 40.0, sand >= 20.0, sand <= 45),
        2.0,  # clay
        np.where(
            np.logical_and(clay >= 27.0, sand >= 20.0, sand <= 45),
            1.7,  # clay loam
            np.where(
                np.logical_and(silt <= 40.0, sand <= 20.0),
                2.0,  # clay
                np.where(
                    np.logical_and(silt > 40.0, clay >= 40.0),
                    1.6,  # silty clay
                    np.where(
                        np.logical_and(clay >= 35.0, sand >= 45.0),
                        1.9,  # sandy clay
                        np.where(
                            np.logical_and(clay >= 27.0, sand < 20.0),
                            1.6,  # silty clay loam
                            np.where(
                                np.logical_and(clay <= 10.0, silt >= 80.0),
                                1.2,  # silt
                                np.where(
                                    (silt >= 50.0),
                                    1.5,  # silt loam
                                    np.where(
                                        np.logical_and(
                                            clay >= 7.0, sand <= 52.0, silt >= 28.0
                                        ),
                                        2.0,  # loam
                                        np.where(
                                            (clay >= 20.0),
                                            2.1,  # sandy clay loam
                                            np.where(
                                                (clay >= (sand - 70)),
                                                2.6,  # sandy loam
                                                np.where(
                                                    (clay >= (2 * sand - 170.0)),
                                                    3.0,  # loamy sand
                                                    np.where(
                                                        np.isnan(clay), np.nan, 1.9
                                                    ),  # sand
                                                ),
                                            ),
                                        ),
                                    ),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
        ),
    )

    return erosK


def UsleK_Renard(clay, silt):
    """
    Determine USLE K factor based on Renard.

    Parameters
    ----------
    clay: float
        sand percentage [%].
    silt: float
        silt percentage [%].

    Returns
    -------
    usleK : float
        soil erodibility factor from the USLE equation [-].

    """
    sand = 100 - (clay + silt)

    Dg = np.exp(
        0.01
        * (clay * math.log(0.001) + silt * math.log(0.025) + sand * math.log(0.999))
    )
    usleK = 0.0034 + 0.0405 * np.exp(-0.5 * ((np.log(Dg) + 1.659) / 0.7101) ** 2)

    return usleK


def UsleK_EPIC(clay, silt, oc):
    """
    Determine USLE K factor based on EPIC formula.

    Parameters
    ----------
    clay: float
        sand percentage [%].
    silt: float
        silt percentage [%].
    oc : float
        organic carbon [%].

    Returns
    -------
    usleK : float
        soil erodibility factor from the USLE equation [-].

    """
    sand = 100 - (clay + silt)
    SN = 1 - (sand) / 100

    usleK = (
        (0.2 + 0.3 * np.exp(-0.0256 * sand * (1 - silt / 100)))
        * (silt / (clay + silt)) ** 0.3
        * (1 - (0.25 * oc) / (oc + np.exp(3.72 - 2.95 * oc)))
        * (1 - (0.7 * SN) / (SN + np.exp(22.9 * SN - 5.31)))
    )
    # Conversion to SI units
    usleK = (1 / 7.594) * usleK

    return usleK


def mean_diameter_soil(clay, silt):
    """
    Determine mean diameter of the soil using Fooladmand et al. (2006).

    Parameters
    ----------
    clay: float
        clay percentage [%].
    silt: float
        silt percentage [%].

    Returns
    -------
    d50 : float
        mean diameter of the soil [mm].
    """
    sand = 100 - (clay + silt)
    sand999 = sand * ((999 - 25) / (1000 - 25))

    vd50 = np.log((1 / (0.01 * (clay + silt)) - 1) / (1 / (0.01 * clay) - 1))
    wd50 = np.log((1 / (0.01 * (clay + silt + sand999)) - 1) / (1 / (0.01 * clay) - 1))
    ad50 = 1 / np.log((25 - 1) / (999 - 1))
    bd50 = ad50 / np.log((25 - 1) / 1)
    cd50 = ad50 * np.log(vd50 / wd50)
    ud50 = (-vd50) ** (1 - bd50) / ((-wd50) ** (-bd50))
    d50 = 1 + (-1 / ud50 * np.log(1 / (1 / (0.01 * clay) - 1))) ** (1 / cd50)  # [um]
    d50 = d50 / 1000  # [mm]

    return d50



"""Soilgrid workflows for Wflow plugin."""

import logging

import numpy as np
import pandas as pd
import xarray as xr
from hydromt.gis import raster_utils
from scipy.optimize import curve_fit

from hydromt_wflow.workflows import ptf, soilparams

logger = logging.getLogger(f"hydromt.{__name__}")

__all__ = [
    "soilgrids",
    "soilgrids_sediment",
    "soilgrids_brooks_corey",
    "update_soil_with_paddy",
]

# soilgrids_2017
soildepth_cm = np.array([0.0, 5.0, 15.0, 30.0, 60.0, 100.0, 200.0])
soildepth_mm = 10.0 * soildepth_cm
# for 2017 - midpoint and midpoint surface are equal to original
soildepth_cm_midpoint = np.array([0.0, 5.0, 15.0, 30.0, 60.0, 100.0, 200.0])
soildepth_mm_midpoint = 10.0 * soildepth_cm
soildepth_mm_midpoint_surface = 10.0 * soildepth_cm
M_minmax = 10000.0
nodata = -9999.0

# default z [cm] of soil layers wflow_sbm: 10, 40, 120, > 120
# mapping of wflow_sbm parameter soil_brooks_corey_c to soil layer SoilGrids
c_sl_index = [2, 4, 6, 7]  # v2017 direct mapping
# c_sl_index = [2, 3, 5, 6] #v2021 if direct mapping - not used,
# averages are taken instead.


def concat_layers(
    ds: xr.Dataset,
    soil_fn: str = "soilgrids",
    variables: list[str] = ["bd", "oc", "ph", "clyppt", "sltppt", "sndppt"],
):
    """
    Preprocess functions to concat soilgrids along a layer dimension.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing soil properties.
    soil_fn : str
        soilgrids version {'soilgrids', 'soilgrids_2020'}
    variables : list
        List of soil properties to concat.
    """
    if soil_fn == "soilgrids_2020":
        nb_sl = 6
    else:
        nb_sl = 7
    ds = ds.assign_coords(sl=np.arange(1, nb_sl + 1))

    for var in variables:
        da_prop = []
        for i in np.arange(1, nb_sl + 1):
            da_prop.append(ds[f"{var}_sl{i}"])
            # remove layer from ds
            ds = ds.drop_vars(f"{var}_sl{i}")
        da = xr.concat(
            da_prop,
            pd.Index(np.arange(1, nb_sl + 1, dtype=int), name="sl"),
        ).transpose("sl", ...)
        da.name = var
        # add concat maps to ds
        ds[f"{var}"] = da

    return ds


def average_soillayers_block(ds, soilthickness):
    """
    Determine weighted average of soil property at different depths over soil thickness.

    Assuming that the properties are computed at the mid-point of the interval and
    are considered constant over the whole depth interval (Sousa et al., 2020).
    https://doi.org/10.5194/soil-2020-65
    This function is used for soilgrids_2020.

    Parameters
    ----------
    ds: xarray.Dataset
        Dataset containing soil property over each soil depth profile [sl1 - sl6].
    soilthickness : xarray.DataArray
        Dataset containing soil thickness [cm].

    Returns
    -------
    da_av : xarray.DataArray
        Dataset containing weighted average of soil property.

    """
    da_sum = soilthickness * 0.0
    # set NaN values to 0.0 (to avoid RuntimeWarning in comparison soildepth)
    d = soilthickness.fillna(0.0)

    for i in ds.sl:
        da_sum = da_sum + (
            (soildepth_cm[i] - soildepth_cm[i - 1])
            * ds.sel(sl=i)
            * (d >= soildepth_cm[i])
            + (d - soildepth_cm[i - 1])
            * ds.sel(sl=i)
            * ((d < soildepth_cm[i]) & (d > soildepth_cm[i - 1]))
        )

    da_av = xr.apply_ufunc(
        lambda x, y: x / y,
        da_sum,
        soilthickness,
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=True,
    )

    da_av.raster.set_nodata(np.nan)

    return da_av


def average_soillayers(ds, soilthickness):
    """
    Determine weighted average of soil property at different depths over soil thickness.

    Using the trapezoidal rule.
    See also: Hengl, T., Mendes de Jesus, J., Heuvelink, G. B. M., Ruiperez Gonzalez,
    M., Kilibarda, M., Blagotic, A., et al.: SoilGrids250m:
    Global gridded soil information based on machine learning,
    PLoS ONE, 12, https://doi.org/10.1371/journal.pone.0169748, 2017.
    This function is used for soilgrids (2017).

    Parameters
    ----------
    ds: xarray.Dataset
        Dataset containing soil property at each soil depth [sl1 - sl7].
    soilthickness : xarray.DataArray
        Dataset containing soil thickness [cm].

    Returns
    -------
    da_av : xarray.DataArray
        Dataset containing the weighted average of the soil property.

    """
    da_sum = soilthickness * 0.0
    # set NaN values to 0.0 (to avoid RuntimeWarning in comparison soildepth)
    d = soilthickness.fillna(0.0)

    for i in range(1, len(ds.sl)):  # range(1, 7):
        da_sum = da_sum + (
            (soildepth_cm[i] - soildepth_cm[i - 1])
            * (ds.sel(sl=i) + ds.sel(sl=i + 1))
            * (d >= soildepth_cm[i])
            + (d - soildepth_cm[i - 1])
            * (ds.sel(sl=i) + ds.sel(sl=i + 1))
            * ((d < soildepth_cm[i]) & (d > soildepth_cm[i - 1]))
        )

    da_av = xr.apply_ufunc(
        lambda x, y: x * (1 / (y * 2)),
        da_sum,
        soilthickness,
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=True,
    )

    da_av.raster.set_nodata(np.nan)

    return da_av


def pore_size_distribution_index_layers(ds, thetas):
    """
    Determine pore size distribution index per soil layer depth based on PTF.

    Parameters
    ----------
    ds: xarray.Dataset
        Dataset containing soil properties at each soil depth [sl1 - sl7].
    thetas: xarray.Dataset
        Dataset containing theta_s at each soil layer depth.

    Returns
    -------
    ds_out : xarray.Dataset
        Dataset containing pore size distribution index [-] for each soil layer
        depth.
    """
    ds_out = xr.apply_ufunc(
        ptf.pore_size_index_brakensiek,
        ds["sndppt"],
        thetas,
        ds["clyppt"],
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=True,
    )
    ds_out.name = "pore_size"
    ds_out.raster.set_nodata(np.nan)
    # ds_out = ds_out.raster.interpolate_na("nearest")
    return ds_out


def brooks_corey_layers(
    thetas_sl: xr.Dataset,
    ds: xr.Dataset,
    ds_like: xr.Dataset,
    soil_fn: str = "soilgrids",
    wflow_layers: list[int] = [100, 300, 800],
    soildepth_cm: np.array = np.array([0.0, 5.0, 15.0, 30.0, 60.0, 100.0, 200.0]),
):
    """
    Determine Brooks Corey coefficient per wflow soil layer depth.

    First pore size distribution index is computed based on theta_s and other soil
    parameters and scaled to the model resolution.

    Then the Brooks Corey coefficient is computed for each wflow soil layer depth
    by weighted averaging the pore size distribution index over the soil thickness.

    Parameters
    ----------
    thetas_sl: xarray.Dataset
        Dataset containing theta_s at each soil layer depth.
    ds: xarray.Dataset
        Dataset containing soil properties at each soil depth.
    ds_like: xarray.Dataset
        Dataset at model resolution for reprojection.
    soil_fn: str
        soilgrids version {'soilgrids', 'soilgrids_2020'}
    wflow_layers: list
        List of soil layer depths [cm] for which soil_brooks_corey_c is calculated.
    soildepth_cm: np.array
        Depth of each soil layers [cm].

    Returns
    -------
    ds_c : xarray.Dataset
        Dataset containing soil_brooks_corey_c for the wflow_sbm soil layers.
    """
    # Get pore size distribution index
    lambda_sl_hr = pore_size_distribution_index_layers(ds, thetas_sl)
    lambda_sl = np.log(lambda_sl_hr)
    lambda_sl = lambda_sl.raster.reproject_like(ds_like, method="average")
    lambda_sl = np.exp(lambda_sl)

    # Brooks Corey coefficient
    c_sl = 3.0 + (2.0 / lambda_sl)
    c_sl.name = "c_sl"

    # Resample for the wflow layers
    wflow_thickness = wflow_layers.copy()
    # Go from wflow layer thickness to soil depths (cumulative)
    wflow_depths = np.cumsum(wflow_thickness)
    # Check if the last wflow depth is less than 2000 mm (soilgrids limit)
    if wflow_depths[-1] > 2000:
        raise ValueError(
            "The total depth of the wflow soil layers should be 2000 mm, as the "
            "soilgrids data does not go deeper than 2 m."
        )
    # Add a zero for the first depth and 2000 for the last depth
    wflow_depths = np.insert(wflow_depths, 0, 0)
    wflow_depths = np.append(wflow_depths, 2000)
    # Compute the thickness of the last layer
    wflow_thickness.append(2000 - wflow_depths[-2])

    # Soil data depth
    soildepth = soildepth_cm * 10

    # make empty dataarray for soil_brooks_corey_c for the 4 sbm layers
    ds_c = raster_utils.full(
        coords=c_sl.raster.coords,
        nodata=np.nan,
        crs=c_sl.raster.crs,
        dtype="float32",
        name="soil_brooks_corey_c",
        attrs={"thickness": wflow_thickness, "soil_fn": soil_fn},
    )
    ds_c = ds_c.expand_dims(dim=dict(layer=np.arange(len(wflow_thickness)))).copy()

    # calc weighted average values of soil_brooks_corey_c over the sbm soil layers
    for nl in range(len(wflow_thickness)):
        top_depth = wflow_depths[nl]
        bottom_depth = wflow_depths[nl + 1]
        c_nl_sum = None
        for d in range(len(soildepth) - 1):
            if soil_fn == "soilgrids_2020":
                c_av = c_sl.sel(sl=d + 1)
            else:
                c_av = (c_sl.sel(sl=d + 1) + c_sl.sel(sl=d + 2)) / 2
            # wflow layer fully within soilgrid layer
            if soildepth[d] <= top_depth and soildepth[d + 1] >= bottom_depth:
                c_nl = c_av * (bottom_depth - top_depth)
            # layer fully within wflow layer
            elif soildepth[d] >= top_depth and soildepth[d + 1] <= bottom_depth:
                c_nl = c_av * (soildepth[d + 1] - soildepth[d])
            # bottom part of the layer within wflow
            elif soildepth[d] <= bottom_depth and soildepth[d + 1] >= bottom_depth:
                c_nl = c_av * (bottom_depth - soildepth[d])
            # top part of the layer within wflow
            elif soildepth[d] <= top_depth and soildepth[d + 1] >= top_depth:
                c_nl = c_av * (soildepth[d + 1] - top_depth)
            # layer outside of wflow layer
            else:
                c_nl = None

            # Add to the sum
            if c_nl is not None:
                if c_nl_sum is None:
                    c_nl_sum = c_nl
                else:
                    c_nl_sum = c_nl_sum + c_nl

        ds_c.loc[dict(layer=nl)] = c_nl_sum / wflow_thickness[nl]

    return ds_c


def kv_layers(ds, thetas, ptf_name):
    """
    Determine vertical saturated hydraulic conductivity per soil layer depth.

    Based on PTF.

    Parameters
    ----------
    ds: xarray.Dataset
        Dataset containing soil properties at each soil depth [sl1 - sl7].
    thetas: xarray.Dataset
        Dataset containing theta_s at each soil layer depth.
    ptf_name : str
        PTF to use for calculation ksat_vertical .

    Returns
    -------
    ds_out : xarray.Dataset
        Dataset containing ksat_vertical  [mm/day] for each soil layer depth.
    """
    if ptf_name == "brakensiek":
        ds_out = xr.apply_ufunc(
            ptf.kv_brakensiek,
            thetas,
            ds["clyppt"],
            ds["sndppt"],
            dask="parallelized",
            output_dtypes=[float],
            keep_attrs=True,
        )
    elif ptf_name == "cosby":
        ds_out = xr.apply_ufunc(
            ptf.kv_cosby,
            ds["clyppt"],
            ds["sndppt"],
            dask="parallelized",
            output_dtypes=[float],
            keep_attrs=True,
        )

    ds_out.name = "kv"
    ds_out.raster.set_nodata(np.nan)

    return ds_out


def func(x, b):
    return np.exp(-b * x)


def do_linalg(x, y):
    """
    Apply np.linalg.lstsq and return fitted parameter.

    Parameters
    ----------
    x : array_like (float)
        "Coefficient"  matrix.
    y : array_like (float)
        dependent variable.

    Returns
    -------
    popt_0 : float
        Optimal value for the parameter fit.

    """
    idx = (~np.isinf(np.log(y))) & (~np.isnan(y))
    return np.linalg.lstsq(x[idx, np.newaxis], np.log(y[idx]), rcond=None)[0][0]


def do_curve_fit(x, y):
    """
    Apply scipy.optimize.curve_fit and return fitted parameter.

    If least-squares minimization fails with an initial guess p0 of 1e-3,
    and 1e-4, np.linalg.lstsq is used for curve fitting.

    Parameters
    ----------
    x : array_like of M length (float)
        independent variable.
    y : array_like of M length (float)
        dependent variable.

    Returns
    -------
    popt_0 : float
        Optimal value for the parameter fit.

    """
    x = np.asarray(x)
    y = np.asarray(y)

    mask = np.isfinite(y) & (y > 0)
    if not np.any(mask):
        return np.nan

    x_valid = x[mask]
    y_valid = y[mask]

    # Attempt nonlinear fit
    for p0 in (1e-3, 1e-4):
        try:
            popt, _ = curve_fit(func, x_valid, y_valid, p0=(p0,))
            return float(popt[0])
        except RuntimeError:
            pass

    # Fallback: linear regression in log-space
    logy = np.log(y_valid)
    coef, *_ = np.linalg.lstsq(x_valid[:, None], logy, rcond=None)

    # lstsq estimates the slope (b) of the line log(y) = -b * x,
    # so we return -coef[0] to get the true b.
    return float(-coef[0])


def constrain_M(M, popt_0, M_minmax):
    """Constrain M parameter with the value M_minmax."""
    M = xr.where((M > 0) & (popt_0 == 0), M_minmax, M)
    M = xr.where(M > M_minmax, M_minmax, M)
    M = xr.where(M < 0, M_minmax, M)
    return M


def soilgrids(
    ds: xr.Dataset,
    ds_like: xr.Dataset,
    ptfKsatVer: str = "brakensiek",
    soil_fn: str = "soilgrids",
    wflow_layers: list[int] = [100, 300, 800],
):
    """
    Return soil parameter maps at model resolution.

    Based on soil properties from SoilGrids datasets.
    Both soilgrids 2017 and 2020 are supported. Soilgrids 2017 provides soil properties
    at 7 specific depths, while soilgrids_2020 provides soil properties averaged over
    6 depth intervals.
    Ref: Hengl, T., Mendes de Jesus, J., Heuvelink, G. B. M., Ruiperez Gonzalez,
    M., Kilibarda, M., Blagotic, A., et al.: SoilGrids250m:
    Global gridded soil information based on machine learning,
    PLoS ONE, 12, https://doi.org/10.1371/journal.pone.0169748, 2017.
    Ref: de Sousa, L.M., Poggio, L., Batjes, N.H., Heuvelink, G., Kempen, B., Riberio,
    E. and Rossiter, D., 2020. SoilGrids 2.0:
    producing quality-assessed soil information for the globe. SOIL Discussions,
    pp.1-37. https://doi.org/10.5194/soil-2020-65.

    A ``soil_mapping`` table can optionally be provided to derive parameters based
    on soil texture classes. A default table *soil_mapping_default* is available
    to derive the infiltration capacity of the soil.

    The following soil parameter maps are calculated:

        - **theta_s** : average saturated soil water content [m3/m3]
        - **theta_r** : average residual water content [m3/m3]
        - **ksat_vertical** : vertical saturated hydraulic conductivity at soil
            surface [mm/day]
        - **soil_thickness** : soil thickness [mm]
        - **f** : scaling parameter controlling the decline of ksat_vertical [mm-1]
            (fitted with curve_fit (scipy.optimize)), bounds are checked
        - **soil_f_** : scaling parameter controlling the decline of ksat_vertical
            [mm-1] (fitted with numpy linalg regression), bounds are checked
        - **soil_brooks_corey_c_** map: Brooks Corey coefficients [-] based on pore
            size distribution index for the wflow_sbm soil layers.
        - **meta_{soil_fn}_ksat_vertical_[z]cm** : ksat vertical [mm/day] at soil
            depths [z] of SoilGrids data [0.0, 5.0, 15.0, 30.0, 60.0, 100.0, 200.0]
        - **meta_soil_texture** : USDA Soil texture based on percentage clay, silt,
            sand mapping: [1:Clay, 2:Silty Clay, 3:Silty Clay-Loam, 4:Sandy Clay,
            5:Sandy Clay-Loam, 6:Clay-Loam, 7:Silt, 8:Silt-Loam, 9:Loam, 10:Sand,
            11: Loamy Sand, 12:Sandy Loam]


    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing soil properties.
    ds_like : xarray.DataArray
        Dataset at model resolution.
    ptfKsatVer : str
        PTF to use for calculation ksat_vertical .
    soil_fn : str
        soilgrids version {'soilgrids', 'soilgrids_2020'}
    wflow_layers : list
        List of soil layer depths [cm] for which soil_brooks_corey_c is calculated.

    Returns
    -------
    ds_out : xarray.Dataset
        Dataset containing gridded soil parameters.
    """
    if soil_fn == "soilgrids_2020":
        # use midpoints of depth intervals for soilgrids_2020.
        soildepth_cm_midpoint = np.array([2.5, 10.0, 22.5, 45.0, 80.0, 150.0])
        soildepth_cm_midpoint_surface = np.array([0, 10.0, 22.5, 45.0, 80.0, 150.0])
        soildepth_cm = np.array([0.0, 5.0, 15.0, 30.0, 60.0, 100.0, 200.0])
        soildepth_mm_midpoint_surface = 10.0 * soildepth_cm_midpoint_surface
    else:
        soildepth_cm = np.array([0.0, 5.0, 15.0, 30.0, 60.0, 100.0, 200.0])
        soildepth_cm_midpoint = np.array([0.0, 5.0, 15.0, 30.0, 60.0, 100.0, 200.0])
        soildepth_mm_midpoint_surface = 10.0 * soildepth_cm

    ds_out = xr.Dataset(coords=ds_like.raster.coords)

    # set nodata values in dataset to NaN (based on soil property SLTPPT at
    # first soil layer)
    # ds = xr.where(ds["sltppt_sl1"] == ds["sltppt_sl1"].raster.nodata, np.nan, ds)
    ds = ds.raster.mask_nodata()

    # concat along a sl dimension
    ds = concat_layers(ds, soil_fn)

    logger.info("calculate and resample theta_s")
    thetas_sl = xr.apply_ufunc(
        ptf.thetas_toth,
        ds["ph"],
        ds["bd"],
        ds["clyppt"],
        ds["sltppt"],
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=True,
    )

    if soil_fn == "soilgrids_2020":
        thetas = average_soillayers_block(thetas_sl, ds["soilthickness"])
    else:
        thetas = average_soillayers(thetas_sl, ds["soilthickness"])
    thetas = thetas.raster.reproject_like(ds_like, method="average")
    ds_out["theta_s"] = thetas.astype(np.float32)

    logger.info("calculate and resample theta_r")
    thetar_sl = xr.apply_ufunc(
        ptf.thetar_rawls_brakensiek,
        ds["sndppt"],
        ds["clyppt"],
        thetas_sl,
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=True,
    )

    if soil_fn == "soilgrids_2020":
        thetar = average_soillayers_block(thetar_sl, ds["soilthickness"])
    else:
        thetar = average_soillayers(thetar_sl, ds["soilthickness"])
    thetar = thetar.raster.reproject_like(ds_like, method="average")
    ds_out["theta_r"] = thetar.astype(np.float32)

    soilthickness_hr = ds["soilthickness"]
    soilthickness = soilthickness_hr.raster.reproject_like(ds_like, method="average")
    # wflow_sbm cannot handle (yet) zero soil thickness
    soilthickness = soilthickness.where(soilthickness > 0.0, np.nan)
    soilthickness.raster.set_nodata(np.nan)
    soilthickness = soilthickness.astype(np.float32)
    ds_out["soil_thickness"] = soilthickness * 10.0  # from [cm] to [mm]

    logger.info("calculate and resample ksat_vertical ")
    kv_sl_hr = kv_layers(ds, thetas_sl, ptfKsatVer)
    kv_sl = np.log(kv_sl_hr)
    kv_sl = kv_sl.raster.reproject_like(ds_like, method="average")
    kv_sl = np.exp(kv_sl)

    logger.info("calculate and resample pore size distribution index")
    ds_c = brooks_corey_layers(
        thetas_sl=thetas_sl,
        ds=ds,
        ds_like=ds_like,
        soil_fn=soil_fn,
        wflow_layers=wflow_layers,
        soildepth_cm=soildepth_cm,
    )
    ds_out["soil_brooks_corey_c"] = ds_c

    ds_out["ksat_vertical"] = kv_sl.sel(sl=1).astype(np.float32)

    for i, sl in enumerate(kv_sl["sl"]):
        kv = kv_sl.sel(sl=sl)
        ds_out[
            f"meta_{soil_fn}_ksat_vertical_" + str(soildepth_cm_midpoint[i]) + "cm"
        ] = kv.astype(np.float32)

    kv = kv_sl / kv_sl.sel(sl=1)
    logger.info(
        "fit z - log(ksat_vertical ) with numpy linalg regression (y = b*x) -> M_"
    )
    popt_0_ = xr.apply_ufunc(
        do_linalg,
        soildepth_mm_midpoint_surface,
        kv.compute(),
        vectorize=True,
        dask="parallelized",
        input_core_dims=[["z"], ["sl"]],
        output_dtypes=[float],
        keep_attrs=True,
    )

    M_ = (thetas - thetar) / (-popt_0_)
    M_ = constrain_M(M_, popt_0_, M_minmax)
    ds_out["soil_f_"] = ((thetas - thetar) / M_).astype(np.float32)

    logger.info("fit zi - Ksat with curve_fit (scipy.optimize) -> M")
    popt_0 = xr.apply_ufunc(
        do_curve_fit,
        soildepth_mm_midpoint_surface,
        kv.compute(),
        vectorize=True,
        dask="parallelized",
        input_core_dims=[["z"], ["sl"]],
        output_dtypes=[float],
        keep_attrs=True,
    )

    M = (thetas - thetar) / (popt_0)
    M = constrain_M(M, popt_0, M_minmax)
    ds_out["f"] = ((thetas - thetar) / M).astype(np.float32)

    # wflow soil map is based on USDA soil classification
    # soilmap = ds["tax_usda"].raster.interpolate_na()
    # soilmap = soilmap.raster.reproject_like(ds_like, method="mode")
    # ds_out["wflow_soil"] = soilmap.astype(np.float32)

    # wflow_soil map is based on soil texture calculated with percentage
    # sand, silt, clay
    # clay, silt percentages averaged over soil thickness
    if soil_fn == "soilgrids_2020":
        clay_av = average_soillayers_block(ds["clyppt"], ds["soilthickness"])
        silt_av = average_soillayers_block(ds["sltppt"], ds["soilthickness"])
    else:
        clay_av = average_soillayers(ds["clyppt"], ds["soilthickness"])
        silt_av = average_soillayers(ds["sltppt"], ds["soilthickness"])

    # calc soil texture
    soil_texture = xr.apply_ufunc(
        ptf.soil_texture_usda,
        clay_av,
        silt_av,
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=True,
    )

    soil_texture_out = soil_texture.raster.reproject_like(ds_like, method="mode")
    # np.nan is not a valid value for array with type integer
    ds_out["meta_soil_texture"] = soil_texture_out
    ds_out["meta_soil_texture"].raster.set_nodata(0)

    dtypes = {"meta_soil_texture": np.int32}
    for var in ds_out:
        dtype = dtypes.get(var, np.float32)
        logger.debug(f"Interpolate nodata (NaN) values for {var}")
        ds_out[var] = ds_out[var].raster.interpolate_na("nearest")
        ds_out[var] = ds_out[var].fillna(nodata).astype(dtype)
        ds_out[var].raster.set_nodata(np.dtype(dtype).type(nodata))

    return ds_out


def soilgrids_brooks_corey(
    ds: xr.Dataset,
    ds_like: xr.Dataset,
    soil_fn: str = "soilgrids",
    wflow_layers: list[int] = [100, 300, 800],
):
    """
    Determine Brooks Corey coefficient per wflow soil layer depth.

    First pore size distribution index is computed based on theta_s and other soil
    parameters and scaled to the model resolution.

    Then the Brooks Corey coefficient is computed for each wflow soil layer depth
    by weighted averaging the pore size distribution index over the soil thickness.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing soil properties.
    ds_like : xarray.DataArray
        Dataset at model resolution.
    soil_fn : str
        soilgrids version {'soilgrids', 'soilgrids_2020'}
    wflow_layers : list
        List of wflow soil layer depths [cm] over which soil_brooks_corey_c is
        calculated.

    Returns
    -------
    ds_out : xarray.Dataset
        Dataset containing soil_brooks_corey_c for the wflow_sbm soil layers.
    """
    if soil_fn == "soilgrids_2020" or soil_fn == "soilgrids":
        soildepth_cm = np.array([0.0, 5.0, 15.0, 30.0, 60.0, 100.0, 200.0])
    else:
        raise ValueError("Only soilgrids_2020 and soilgrids are supported.")

    ds_out = xr.Dataset(coords=ds_like.raster.coords)
    ds = ds.raster.mask_nodata()
    # concat along a sl dimension
    ds = concat_layers(ds, soil_fn)

    logger.info("calculate and resample theta_s")
    thetas_sl = xr.apply_ufunc(
        ptf.thetas_toth,
        ds["ph"],
        ds["bd"],
        ds["clyppt"],
        ds["sltppt"],
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=True,
    )

    logger.info("calculate and resample pore size distribution index")
    # Brooks Corey coefficient
    ds_c = brooks_corey_layers(
        thetas_sl=thetas_sl,
        ds=ds,
        ds_like=ds_like,
        soil_fn=soil_fn,
        wflow_layers=wflow_layers,
        soildepth_cm=soildepth_cm,
    )
    ds_out["soil_brooks_corey_c"] = ds_c

    return ds_out


def soilgrids_sediment(
    ds: xr.Dataset,
    ds_like: xr.Dataset,
    usle_k_method: str = "renard",
    add_aggregates: bool = True,
) -> xr.Dataset:
    """
    Return soil parameter maps for sediment modelling at model resolution.

    Based on soil properties from SoilGrids dataset.
    Sediment size distribution and addition of small and large aggregates can be
    estimated from primary particle size distribution with Foster et al. (1980).
    USLE K factor can be computed from the soil data using Renard or EPIC methods.
    Calculation of D50 and fraction of fine and very fine sand (fvfs) from
    Fooladmand et al, 2006.

    The following soil parameter maps are calculated:

        * soil_clay_fraction: clay content of the topsoil [g/g]
        * soil_silt_fraction: silt content of the topsoil [g/g]
        * soil_sand_fraction: sand content of the topsoil [g/g]
        * soil_sagg_fraction: small aggregate content of the topsoil [g/g]
        * soil_lagg_fraction: large aggregate content of the topsoil [g/g]
        * erosion_soil_detachability: mean detachability of the soil
            (Morgan et al., 1998) [g/J]
        * usle_k: soil erodibility factor from the USLE equation [-]
        * soil_sediment_d50: median sediment diameter of the soil [mm]
        * land_govers_c: Govers factor for overland flow transport capacity [-]
        * land_govers_n: Govers exponent for overland flow transport capacity [-]

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing soil properties.
    ds_like : xarray.DataArray
        Dataset at model resolution.
    usle_k_method : str
        Method to use for calculation of USLE_K {"renard", "epic"}.
    add_aggregates : bool
        Add small and large aggregates to the soil properties. Default is True.

    Returns
    -------
    ds_out : xarray.Dataset
        Dataset containing gridded soil parameters for sediment modelling.
    """
    ds_out = xr.Dataset(coords=ds_like.raster.coords)

    # set nodata values in dataset to NaN (based on soil property SLTPPT at
    # first soil layer)
    # ds = xr.where(ds["sltppt_sl1"] == ds["sltppt_sl1"].raster.nodata, np.nan, ds)
    ds = ds.raster.mask_nodata()

    # soil properties
    pclay = ds["clyppt_sl1"]
    psilt = ds["sltppt_sl1"]
    poc = ds["oc_sl1"]

    # Detachability of the soil
    erosK = xr.apply_ufunc(
        ptf.ErosK_texture,
        pclay,
        psilt,
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=True,
    )
    erosK = erosK.raster.reproject_like(ds_like, method="average")
    ds_out["erosion_soil_detachability"] = erosK.astype(np.float32)

    # USLE K parameter
    if usle_k_method == "renard":
        usle_k = xr.apply_ufunc(
            ptf.UsleK_Renard,
            pclay,
            psilt,
            dask="parallelized",
            output_dtypes=[float],
            keep_attrs=True,
        )
    elif usle_k_method == "epic":
        usle_k = xr.apply_ufunc(
            ptf.UsleK_EPIC,
            pclay,
            psilt,
            poc,
            dask="parallelized",
            output_dtypes=[float],
            keep_attrs=True,
        )
    usle_k = usle_k.raster.reproject_like(ds_like, method="average")
    ds_out["usle_k"] = usle_k.astype(np.float32)

    # Mean diameter of the soil
    d50 = xr.apply_ufunc(
        ptf.mean_diameter_soil,
        pclay,
        psilt,
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=True,
    )
    ds_out["soil_sediment_d50"] = d50.raster.reproject_like(
        ds_like, method="average"
    ).astype(np.float32)

    # Govers factor and exponent for overland flow transport capacity
    c_govers = ((d50 * 1000 + 5) / 0.32) ** (-0.6)
    n_govers = ((d50 * 1000 + 5) / 300) ** (0.25)
    ds_out["land_govers_c"] = c_govers.raster.reproject_like(
        ds_like, method="average"
    ).astype(np.float32)
    ds_out["land_govers_n"] = n_govers.raster.reproject_like(
        ds_like, method="average"
    ).astype(np.float32)

    # Sediment size distribution
    if add_aggregates:
        fclay = 0.20 * pclay / 100
        fsilt = 0.13 * psilt / 100
        fsand = (1 - pclay / 100 - psilt / 100) * (1 - pclay / 100) ** 2.4
        fsagg = 0.28 * (pclay / 100 - 0.25) + 0.5
        fsagg = fsagg.where(pclay < 50, 0.57)
        fsagg = fsagg.where(pclay > 25, 2 * pclay / 100)
    else:
        fclay = pclay / 100
        fsilt = psilt / 100
        fsagg = fclay * 0.0

    # Reproject to model resolution
    fclay = fclay.raster.reproject_like(ds_like, method="average")
    ds_out["soil_clay_fraction"] = fclay.astype(np.float32)
    fsilt = fsilt.raster.reproject_like(ds_like, method="average")
    ds_out["soil_silt_fraction"] = fsilt.astype(np.float32)
    fsagg = fsagg.raster.reproject_like(ds_like, method="average")
    ds_out["soil_sagg_fraction"] = fsagg.astype(np.float32)
    # Make sure that the sum of the percentages is 100
    if add_aggregates:
        fsand = fsand.raster.reproject_like(ds_like, method="average").astype(
            np.float32
        )
        ds_out["soil_sand_fraction"] = fsand
        ds_out["soil_lagg_fraction"] = (1 - fclay - fsilt - fsand - fsagg).astype(
            np.float32
        )
    else:
        ds_out["soil_sand_fraction"] = (1 - fclay - fsilt).astype(np.float32)
        ds_out["soil_lagg_fraction"] = (fclay * 0.0).astype(np.float32)

    for var in ds_out:
        ds_out[var] = ds_out[var].raster.interpolate_na("nearest")
        logger.info(f"Interpolate NAN values for {var}")
        ds_out[var] = ds_out[var].fillna(nodata)
        ds_out[var].raster.set_nodata(nodata)

    return ds_out


def update_soil_with_paddy(
    ds: xr.Dataset,
    ds_like: xr.Dataset,
    paddy_mask: xr.DataArray,
    soil_fn: str = "soilgrids",
    update_c: bool = True,
    wflow_layers: list[int] = [50, 100, 50, 200, 800],
    target_conductivity: list[None | int | float] = [None, None, 5, None, None],
):
    """
    Update soil_brooks_corey_c and soil_ksat_vertical_factor for paddy fields.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing soil properties.
    ds_like : xarray.DataArray
        Dataset at model resolution.
        Required variables: basins, ksat_vertical, f, elevtn, soil_brooks_corey_c
    paddy_mask : xarray.DataArray
        Dataset containing paddy fields mask.
    soil_fn : str
        soilgrids version {'soilgrids', 'soilgrids_2020'}
    update_c : bool
        Update soil_brooks_corey_c based on change in wflow_layers.
    wflow_layers : list
        List of soil layer depths [cm] for which soil_brooks_corey_c is calculated.
    target_conductivity : list
        Target conductivity for each wflow layer.

    Returns
    -------
    soil_out : xarray.Dataset
        Dataset containing updated soil properties.
    """
    if len(wflow_layers) != len(target_conductivity):
        raise ValueError(
            "Lengths of wflow_thicknesslayers and target_conductivity does not match"
        )

    # Set soil_ksat_vertical_factor maps
    # Determine the fraction required to reach target_conductivity
    # Using to wflow exponential decline to determine the conductivity at the
    # required depth Find value at the bottom of the required layer and infer
    # required correction factor for that layer Values are only set for locations
    # with paddy irrigation, all other cells are set to be equal to 1
    logger.info("Adding soil_ksat_vertical_factor map")
    kv0 = ds_like["ksat_vertical"]
    f = ds_like["f"]
    kv0_mask = kv0.where(paddy_mask == 1)
    f_mask = f.where(paddy_mask == 1)

    # update soil_brooks_corey_c
    if update_c:
        ds_out = soilgrids_brooks_corey(
            ds=ds,
            ds_like=ds_like,
            soil_fn=soil_fn,
            wflow_layers=wflow_layers,
        )
        for var in ds_out:
            dtype = np.float32
            logger.debug(f"Interpolate nodata (NaN) values for {var}")
            ds_out[var] = ds_out[var].raster.interpolate_na("nearest")
            ds_out[var] = ds_out[var].where(
                ~ds_like["basins"].raster.mask_nodata().isnull()
            )
            ds_out[var] = ds_out[var].fillna(-9999).astype(dtype)
            ds_out[var].raster.set_nodata(np.dtype(dtype).type(-9999))

        # temporarily add the dem to the dataset
        ds_out["elevtn"] = ds_like["elevtn"]

    # Compute the kv_frac (should be done after updating soil_brooks_corey_c!)
    da_kvfrac = soilparams.update_kvfrac(
        ds_model=ds_like if not update_c else ds_out,
        kv0_mask=kv0_mask,
        f_mask=f_mask,
        wflow_thicknesslayers=wflow_layers,
        target_conductivity=target_conductivity,
    )
    if update_c:
        ds_out["soil_ksat_vertical_factor"] = da_kvfrac
        # Remove elevation variable
        ds_out = ds_out.drop_vars("elevtn")
    else:
        ds_out = da_kvfrac.to_dataset(name="soil_ksat_vertical_factor")

    return ds_out



"""Glaciers workflows for Wflow plugin."""

import logging
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

logger = logging.getLogger(f"hydromt.{__name__}")

if TYPE_CHECKING:
    import geopandas as gpd
    import xarray as xr

__all__ = ["glaciermaps", "glacierattrs"]


def glaciermaps(
    gdf: "gpd.GeoDataFrame",
    ds_like: "xr.Dataset",
    id_column: str = "simple_id",
    elevtn_name: str = "elevtn",
):
    """Return glacier maps (see list below) at model resolution.

    The following glacier maps are calculated:

    - glacareas: glacier IDs [ID]
    - glacfrac: area fraction of glacier per cell [-]
    - glacstore: storage (volume) of glacier per cell [mm]

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        GeoDataFrame containing glacier geometries and attributes.
    ds_like : xarray.Dataset
        Dataset at model resolution.
    id_column : str, optional, one of "simple_id", "C3S_id", "RGI_id", or "GLIMS_id"
        Column used for the glacier IDs, see data/data_sources.yml.

    Returns
    -------
    ds_out : xarray.Dataset
        Dataset containing gridded glacier data
    """
    # Rasterize the GeoDataFrame to get the areas mask of glaciers with their ids
    ds_out = ds_like.raster.rasterize(
        gdf,
        col_name=id_column,
        nodata=0,
        all_touched=True,
        dtype=None,
        sindex=False,
    ).astype("int32")
    ds_out = ds_out.rename("glacareas")
    ds_out = ds_out.to_dataset()

    # Calculate glacier storage for each glacier
    # 'old' version to calculate storage, by calculating area from the glacier geometry
    # proj = partial(pyproj.transform, pyproj.Proj('epsg:4326'),
    # pyproj.Proj('epsg:3857'))
    # def calcStore(polygon, k=0.2055, gamma=1.375, proj=proj):
    #    return np.int(k * transform(proj, polygon).area ** gamma)
    # gdf['glaciervolume'] = gdf.apply(lambda row: calcStore(row['geometry']), axis=1)
    # def convStoremm(polygon, volm3, rhoice=0.9, proj=proj):
    #    return np.int(rhoice * volm3 / transform(proj, polygon).area * 1000)
    # gdf['glacierstore'] = gdf.apply(lambda row: convStoremm(row['geometry'],
    # row['glaciervolume']), axis=1)
    # 'new' version to calculate storage, because the 'old' version is now returning
    # inf/nan values, using the area listed in the dataset

    gdf["geometry"] = gdf.geometry.buffer(0)  # fix potential geometry errors
    # TODO: use alternative to calculate area? There seems to be a factor 2 difference
    # with the AREA column
    # gdf["AREA2"] = gdf.to_crs(3857).area / 1e6  # km2, area calculation
    # needs projected crs

    def calcStore(areakm2, k=0.2055, gamma=1.375):
        return int(k * (areakm2 * 1e6) ** gamma)

    def convStoremm(areakm2, volm3, rhoice=0.9):
        return int(rhoice * volm3 / (areakm2 * 1e6) * 1000)

    gdf["glaciervolume"] = gdf.apply(lambda row: calcStore(row["AREA"]), axis=1)
    gdf["glacierstore"] = gdf.apply(
        lambda row: convStoremm(row["AREA"], row["glaciervolume"]), axis=1
    )

    # Create vector grid (for calculating fraction and storage per grid cell)
    logger.debug(
        "Creating vector grid for calculating glacier fraction and \
storage per grid cell"
    )
    elevtn = ds_like[elevtn_name]
    idx_valid = np.nonzero(elevtn.values.flatten() != elevtn.raster.nodata)[0]
    gdf_grid = ds_like.raster.vector_grid().loc[idx_valid]
    gdf_grid["glacierfrac"] = np.zeros(len(idx_valid), dtype=np.float32)
    gdf_grid["glacierstore"] = np.zeros(len(idx_valid), dtype=np.float32)
    gdf_grid["area"] = gdf_grid.to_crs(3857).area  # area calculation in projected crs

    # reproject to match ds_like CRS
    gdf = gdf.to_crs(ds_like.raster.crs)

    # Calculate fraction and storage (i.e. volume) per (vector) grid cell
    # Looping over each vector GLACIER
    logger.debug("Setting glacierfrac and store values per glacier.")
    for i in range(len(gdf)):
        glacier = gdf.iloc[i]
        gridded_glacier = gdf_grid.intersection(glacier.geometry)
        gridded_glacier = gridded_glacier.loc[~gridded_glacier.is_empty]
        idxs = gridded_glacier.index
        # logger.debug(f"index: {i}, ID: {glacier[id_column]} // {len(idxs)} cells")
        if np.any(idxs):
            garea_cell = gridded_glacier.to_crs(
                3857
            ).area  # area calculation needs projected crs
            gfrac = garea_cell / np.sum(garea_cell)
            gdf_grid.loc[idxs, "glacierfrac"] += (
                garea_cell / gdf_grid.loc[idxs, "area"]
            ).astype(np.float32)
            gdf_grid.loc[idxs, "glacierstore"] += (
                gfrac * glacier["glacierstore"]
            ).astype(np.float32)

    # reproject back to original projection
    # Create the rasterized glacier storage map
    ds_out["glacstore"] = ds_like.raster.rasterize(
        gdf_grid,
        col_name="glacierstore",
        nodata=0,
        all_touched=False,
        dtype=None,
        sindex=False,
    ).astype("float32")

    # Create the rasterized glacier fraction map
    ds_out["glacfracs"] = ds_like.raster.rasterize(
        gdf_grid,
        col_name="glacierfrac",
        nodata=0,
        all_touched=False,
        dtype=None,
        sindex=False,
    ).astype("float32")

    ds_out["glacareas"].raster.set_nodata(0)
    ds_out["glacstore"].raster.set_nodata(0)
    ds_out["glacfracs"].raster.set_nodata(0)
    return ds_out


def glacierattrs(
    gdf: "gpd.GeoDataFrame",
    TT: float = 1.3,
    Cfmax: float = 5.3,
    SIfrac: float = 0.002,
    id_column: str = "simple_id",
):
    """Return glacier intbls (see list below).

    The following glacier intbls are calculated:

    - glacTempThresh: glacier temperature threshold [°C]
    - glacCfmax: glacier melting factor [mm/(°C*day)]
    - glacSIfrac: fraction of snowpack converted into ice and added to \
glacier storage [-]

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        GeoDataFrame containing glacier geometries and attributes.
    TT : float, optional
        Default value for glacier temperature threshold.
    Cfmax : float, optional
        Default value for glacier melting factor.
    SIfrac : float, optional
        Default value for fraction of snowpack converted into ice and added to \
glacier storage.
    id_column : str, optional, one of "simple_id", "C3S_id", "RGI_id", or "GLIMS_id"
        Column used for the glacier IDs, see data/data_sources.yml.

    Returns
    -------
    df_out : pandas.DataFrame
        DataFrame containing glacier attributes.
    """
    # Initialize output DataFrame with empty values and glacier ID
    df_out = pd.DataFrame(
        index=range(len(gdf[id_column])),
        columns=(
            [
                "glacId",
                "glacTempThresh",
                "glacCfmax",
                "glacSIfrac",
            ]
        ),
    )
    df_out["glacId"] = gdf[id_column].values
    # Fill in other attributes
    df_out["glacTempThresh"] = list(np.full(len(gdf.index), TT))
    df_out["glacCfmax"] = list(np.full(len(gdf.index), Cfmax))
    df_out["glacSIfrac"] = list(np.full(len(gdf.index), SIfrac))

    return df_out



"""Wflow model components submodule."""

from hydromt_wflow.components.config import WflowConfigComponent
from hydromt_wflow.components.forcing import WflowForcingComponent
from hydromt_wflow.components.geoms import WflowGeomsComponent
from hydromt_wflow.components.output_csv import WflowOutputCsvComponent
from hydromt_wflow.components.output_grid import WflowOutputGridComponent
from hydromt_wflow.components.output_scalar import WflowOutputScalarComponent
from hydromt_wflow.components.states import WflowStatesComponent
from hydromt_wflow.components.staticmaps import WflowStaticmapsComponent
from hydromt_wflow.components.tables import WflowTablesComponent

__all__ = [
    "WflowConfigComponent",
    "WflowForcingComponent",
    "WflowGeomsComponent",
    "WflowStatesComponent",
    "WflowStaticmapsComponent",
    "WflowTablesComponent",
    "WflowOutputGridComponent",
    "WflowOutputScalarComponent",
    "WflowOutputCsvComponent",
]



import numpy as np
import xarray as xr

from hydromt_wflow.workflows.reservoirs import set_rating_curve_layer_data_type


def test_set_rating_curve_layer_data_type():
    # create some test data
    data = np.random.randint(4, 5)
    coords = {"x": np.arange(4), "y": np.arange(5)}
    attrs = {"description": "Test data", "_FillValue": -999.0}
    da = xr.DataArray(data, coords=coords, dims=["x", "y"], attrs=attrs).astype(
        np.float32
    )
    da[0, 0] = np.nan
    da2 = da.copy(deep=True)
    da2 = da2.fillna(-999.0)
    ds = xr.Dataset({"reservoir_rating_curve": da})
    ds2 = xr.Dataset({"reservoir_rating_curve": da2})

    ds = set_rating_curve_layer_data_type(ds)
    ds2 = set_rating_curve_layer_data_type(ds2)

    assert ds["reservoir_rating_curve"].dtype == np.int32
    assert ds2["reservoir_rating_curve"].dtype == np.int32
    assert ds["reservoir_rating_curve"].raster.nodata == -999
    assert ds2["reservoir_rating_curve"].attrs["_FillValue"] == -999
    assert ds["reservoir_rating_curve"][0, 0].values == -999



"""Test plugin model class against hydromt.models.model_api."""

from os.path import abspath, dirname, join
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from hydromt_wflow import DATA_DIR
from hydromt_wflow.wflow_base import WflowBaseModel
from hydromt_wflow.wflow_sbm import WflowSbmModel
from hydromt_wflow.wflow_sediment import WflowSedimentModel

TESTDATADIR = join(dirname(abspath(__file__)), "data")
EXAMPLEDIR = join(dirname(abspath(__file__)), "..", "examples")
pytestmark = pytest.mark.integration  # all tests in this module are integration tests

_supported_models: dict[str, type[WflowBaseModel]] = {
    "wflow": WflowSbmModel,
    "wflow_sediment": WflowSedimentModel,
    "wflow_simple": WflowSbmModel,
}


def _assert_dataset_not_empty(ds: xr.Dataset) -> tuple[bool, list[str]]:
    counts = ds.count().to_pandas()
    empty_layers = counts[counts == 0]
    if len(empty_layers) > 0:
        return False, list(counts.index)
    else:
        return True, []


def _assert_grids_not_empty(model: WflowBaseModel):
    if model.staticmaps._data:
        eq, empty_layers = _assert_dataset_not_empty(model.staticmaps._data)
        assert eq, f"empty layers in staticmaps: {empty_layers}"

    if model.forcing._data:
        eq, empty_layers = _assert_dataset_not_empty(model.forcing._data)
        assert eq, f"empty layers in forcing: {empty_layers}"


def _compare_wflow_models(mod0: WflowBaseModel, mod1: WflowBaseModel):
    # check maps
    if mod0.staticmaps._data:
        eq, errors = mod0.staticmaps.test_equal(mod1.staticmaps)
        assert eq, f"staticmaps not equal: {errors}"

    # check geoms
    if mod0.geoms._data:
        eq, errors = mod0.geoms.test_equal(mod1.geoms)
        assert eq, f"geoms not equal: {errors}"

    if mod0.forcing._data:
        # flatten
        eq, errors = mod0.forcing.test_equal(mod1.forcing)
        assert eq, f"forcing not equal: {errors}"

    # check config
    if mod0.config._data:
        # flatten
        eq, errors = mod0.config.test_equal(mod1.config)
        assert eq, f"config not equal: {errors}"


@pytest.mark.timeout(300)  # max 5 min
@pytest.mark.parametrize("model", list(_supported_models.keys()))
@pytest.mark.integration
def test_model_build(tmpdir, model, example_models, example_inis):
    # get model type
    model_type = _supported_models[model]
    # create folder to store new model
    root = str(tmpdir.join(model))
    param_path = DATA_DIR / "parameters_data.yml"
    mod1 = model_type(
        root=root, mode="w", data_libs=["artifact_data", param_path.as_posix()]
    )

    # get ini file
    steps = example_inis[model]

    # Build model
    mod1.build(steps=steps)

    # Compare with model from examples folder
    # (need to read it again for proper geoms check)
    mod1 = model_type(root=root, mode="r")
    mod1.read()
    # get reference model
    mod0 = example_models[model]
    if mod0 is not None:
        mod0.read()

        # make sure models aren't empty
        _assert_grids_not_empty(mod0)

        # compare models
        _compare_wflow_models(mod0, mod1)


def test_base_model_init_should_raise():
    with pytest.raises(TypeError) as exc_info:
        WflowBaseModel()

    assert str(exc_info.value) == (
        "``WflowBaseModel`` is an abstract class and cannot be instantiated "
        "directly. Please use one of its subclasses defined as hydromt-entry "
        "points: [``WflowSbmModel``, ``WflowSedimentModel``]"
    )


@pytest.mark.timeout(60)  # max 1 min
@pytest.mark.integration
def test_model_clip(
    tmpdir: Path,
    example_wflow_model: WflowSbmModel,
    clipped_wflow_model: WflowSbmModel,
):
    model = "wflow"

    # Clip method options
    destination = str(tmpdir.join(model))
    region = {
        "subbasin": [12.3006, 46.4324],
        "meta_streamorder": 4,
    }

    # Clip workflow, based on example model
    example_wflow_model.read()
    example_wflow_model.root.set(destination, mode="w")
    example_wflow_model.clip(region)
    example_wflow_model.write()

    # Compare with model from examples folder
    # (need to read it again for proper geoms check)
    mod1 = WflowSbmModel(root=destination, mode="r")
    mod1.read()
    # Read reference clipped model
    clipped_wflow_model.read()
    # compare models
    _compare_wflow_models(clipped_wflow_model, mod1)
    # check states
    eq, errors = clipped_wflow_model.states.test_equal(mod1.states)
    assert eq, f"states not equal: {errors}"


def test_model_clip_reservoir(
    tmpdir: Path,
    example_wflow_model: WflowSbmModel,
    reservoir_rating: dict[str, pd.DataFrame],
):
    model = "wflow"

    # Clip method options
    destination = str(tmpdir.join(model))
    region = {
        "subbasin": [12.3162, 46.1676],
        "meta_streamorder": 3,
    }

    # Read and add reservoir rating tables before clipping
    clip_model = WflowSbmModel(root=example_wflow_model.root.path, mode="r+")
    clip_model.read()

    for name, tbl in reservoir_rating.items():
        clip_model.tables.set(tbl, name=name)
    assert len(clip_model.tables.data) == 4
    assert "reservoir_sh_169986" in clip_model.tables.data
    assert "reservoir_hq_3367" in clip_model.tables.data

    # Clip
    clip_model.root.set(destination, mode="w")
    clip_model.clip(region)
    assert len(clip_model.tables.data) == 2
    assert "reservoir_sh_169986" in clip_model.tables.data
    assert "reservoir_hq_3367" not in clip_model.tables.data


def test_sediment_model_clip(
    tmpdir: Path,
    example_sediment_model: WflowSedimentModel,
):
    model = "wflow"

    # Clip method options
    destination = str(tmpdir.join(model))
    region = {
        "subbasin": [12.3162, 46.1676],
        "meta_streamorder": 3,
    }

    # Clip workflow, based on example model
    example_sediment_model.read()
    example_sediment_model.root.set(destination, mode="w")
    example_sediment_model.clip(region)

    # Check extent of the clipped model
    n_pixels = example_sediment_model.staticmaps.data["subcatchment"].sum()
    assert n_pixels == 97
    # This model should still have one reservoir after clipping
    assert "reservoirs" in example_sediment_model.geoms.data
    assert len(example_sediment_model.geoms.data["reservoirs"]) == 1
    assert "reservoir_area_id" in example_sediment_model.staticmaps.data
    assert example_sediment_model.config.get_value("model.reservoir__flag") is True


def test_model_inverse_clip(example_wflow_model: WflowSbmModel):
    # Clip method options
    region = {
        "subbasin": [12.3006, 46.4324],
        "meta_streamorder": 4,
    }

    inverse_clip_model = WflowSbmModel(root=example_wflow_model.root.path, mode="r+")
    # Clip workflow, based on example model
    inverse_clip_model.read()
    # Get number of active pixels from full model
    n_pixels_full = inverse_clip_model.staticmaps.data["subcatchment"].sum()
    inverse_clip_model.clip(region.copy(), inverse_clip=True)
    # Get number of active pixels from inversely clipped model
    n_pixels_inverse_clipped = inverse_clip_model.staticmaps.data["subcatchment"].sum()

    # Do clipping again, but normally
    clip_model = WflowSbmModel(root=example_wflow_model.root.path, mode="r+")
    clip_model.read()
    clip_model.clip(region.copy(), inverse_clip=False)
    # Get number of active pixels from clipped model
    n_pixels_clipped = clip_model.staticmaps.data["subcatchment"].sum()

    assert n_pixels_inverse_clipped < n_pixels_full
    assert n_pixels_full == n_pixels_inverse_clipped + n_pixels_clipped


def test_model_outputs(example_wflow_outputs):
    # Tests on outputs
    wflow = example_wflow_outputs

    # Check the gridded output
    assert isinstance(wflow.output_grid.data, xr.Dataset)
    assert "q_river" in wflow.output_grid.data
    assert wflow.output_grid.data.raster.x_dim == wflow.staticmaps.data.raster.x_dim

    # Check the netcdf scalar output
    assert isinstance(wflow.output_scalar.data, xr.Dataset)
    assert "Q" in wflow.output_scalar.data

    # Check the csv output
    assert len(wflow.output_csv.data) == len(
        wflow.config.get_value("output.csv.column")
    )

    # Checks for the csv columns
    # Q for gauges_grdc
    assert len(wflow.output_csv.data["river_q_gauges_grdc"].index) == 3
    assert np.isin(6349410, wflow.output_csv.data["river_q_gauges_grdc"].index)

    # Coordinates and values for coordinate.x and index.x for temp
    assert np.isclose(
        wflow.output_csv.data["temp_bycoord"]["x"].values,
        wflow.output_csv.data["temp_byindex"]["x"].values,
    )
    assert np.allclose(
        wflow.output_csv.data["temp_bycoord"].values,
        wflow.output_csv.data["temp_byindex"].values,
    )

    # Coordinates of the reservoir
    assert np.isclose(wflow.output_csv.data["reservoir_volume"]["y"], 46.16656)



import logging
from collections.abc import Callable
from pathlib import Path
from unittest.mock import MagicMock

import pytest
import xarray as xr

from hydromt_wflow.components import WflowOutputGridComponent
from hydromt_wflow.wflow_sbm import WflowSbmModel


@pytest.fixture
def mock_model(mock_model_factory) -> WflowSbmModel:
    """Fixture to create a mock WflowSbmModel."""
    return mock_model_factory(mode="w")


# Define the side effect function
def get_value_side_effect(key, fallback=None):
    if key == "dir_output":
        return "run_results"
    elif key == "output.netcdf_grid.path":
        return "output.nc"
    return fallback


def test_wflow_output_grid_component_init(mock_model: WflowSbmModel):
    # Setup the component
    component = WflowOutputGridComponent(mock_model)

    # Assert that the internal data is None
    assert component._data is None

    # When asking for data property, it should return a xarray.Dataset
    assert isinstance(component.data, xr.Dataset)
    assert isinstance(component._data, xr.Dataset)  # Same for internal
    assert len(component.data) == 0


def test_wflow_output_grid_component_init_with_region(
    mock_model_staticmaps: WflowSbmModel,
):
    # Setup the component with a region component
    component = WflowOutputGridComponent(
        mock_model_staticmaps, region_component="staticmaps"
    )

    # Check if the region component is set correctly
    assert component._region_component == "staticmaps"
    assert "basin" in component._get_grid_data()


def test_wflow_output_grid_component_set(
    mock_model_staticmaps: WflowSbmModel, forcing_layer: xr.DataArray
):
    # Setup the component
    component = WflowOutputGridComponent(
        mock_model_staticmaps, region_component="staticmaps"
    )

    # Assert that the internal data is None, initializing will happen in set
    assert component._data is None

    # Set an entry
    # Copy else the name gets altered for the next test?
    component.set(forcing_layer.copy(), name="test_layer")

    # Assert the content
    assert isinstance(component._data, xr.Dataset)
    assert "test_layer" in component.data
    assert len(component.data) == 1


def test_wflow_output_grid_component_set_errors(
    mock_model_staticmaps: WflowSbmModel, forcing_layer: xr.DataArray
):
    # Setup the component
    component = WflowOutputGridComponent(
        mock_model_staticmaps, region_component="staticmaps"
    )

    # Try with np.ndarray
    array = forcing_layer.values
    with pytest.raises(TypeError, match="Data must be an xarray Dataset or DataArray"):
        component.set(array)

    # Try with a different grid that is not identical
    # This should raise an error
    different_grid = forcing_layer.copy()
    different_grid["lon"] = different_grid.lon + 1  # Shift x coordinates
    with pytest.raises(
        ValueError, match="Data grid must be identical to staticmaps component"
    ):
        component.set(different_grid)


def test_wflow_output_grid_component_read(
    mock_model_factory: Callable[[Path, str], WflowSbmModel],
    model_subbasin_cached: Path,
):
    # Set it to read mode
    mock_model = mock_model_factory(path=model_subbasin_cached, mode="r")

    # Setup the component
    component = WflowOutputGridComponent(mock_model)

    # Assert its data currently none
    assert component._data is None

    # Read the data
    type(component.model.config).get_value = MagicMock(
        side_effect=get_value_side_effect
    )
    component.read()

    # Assert the read data
    assert isinstance(component.data, xr.Dataset)
    assert len(component.data) == 2
    assert "q_river" in component.data


def test_wflow_output_grid_component_read_init(
    mock_model_factory: Callable[[Path, str], WflowSbmModel],
    model_subbasin_cached: Path,
):
    # Set it to read mode
    mock_model = mock_model_factory(path=model_subbasin_cached, mode="r")

    # Setup the component
    component = WflowOutputGridComponent(mock_model)
    type(component.model.config).get_value = MagicMock(
        side_effect=get_value_side_effect
    )
    assert component._data is None  # Assert no data or structure yet

    # Read at init
    assert len(component.data) == 2
    assert "q_river" in component.data


def test_wflow_output_grid_component_write(
    mock_model: WflowSbmModel,
    grid_dummy_data: xr.DataArray,
    caplog: pytest.LogCaptureFixture,
):
    # Setup the component
    component = WflowOutputGridComponent(mock_model)
    component._data = grid_dummy_data.to_dataset(name="test_layer")

    # Write to a file
    # Check the logger message when calling component.write()
    with caplog.at_level(logging.INFO):
        component.write()
    assert "netcdf_grid is an output of Wflow and will not be written." in caplog.text



"""Wflow forcing component."""

import logging
from datetime import datetime
from pathlib import Path
from typing import Optional

import numpy as np
import pandas as pd
import xarray as xr
from hydromt import hydromt_step
from hydromt.model import Model
from hydromt.model.components import GridComponent, ModelComponent
from hydromt.writers import write_nc

from hydromt_wflow.components import utils

__all__ = ["WflowForcingComponent"]

logger = logging.getLogger(f"hydromt.{__name__}")

_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"


class WflowForcingComponent(GridComponent):
    """Wflow forcing component.

    This class is used for setting, creating, writing, and reading the Wflow forcing.
    The forcing component data stored in the ``data`` property of this class is of the
    hydromt.gis.raster.RasterDataset type which is an extension of xarray.Dataset for
    regular grid.
    """

    def __init__(
        self,
        model: Model,
        *,
        filename: str = "inmaps.nc",
        region_component: str | None = None,
        region_filename: str = "staticgeoms/forcing_region.geojson",
    ):
        """
        Initialize a WflowForcingComponent.

        Parameters
        ----------
        model : Model
            HydroMT model instance
        filename : str, optional
            Default path relative to the root where the forcing file(s) will be read
            and written. By default 'inmaps.nc'.
        region_component : str, optional
            The name of the region component to use as reference for this component's
            region. If None, the region will be set to the grid extent.
        region_filename : str, optional
            A path relative to the root where the region file will be written.
            By default 'staticgeoms/forcing_region.geojson'.
        """
        super().__init__(
            model=model,
            filename=filename,
            region_component=region_component,
            region_filename=region_filename,
        )

    ## I/O methods
    @hydromt_step
    def read(
        self,
        **kwargs,
    ):
        """Read forcing model data at root/dir_input/filename.

        Checks the path of the file in the config toml using both ``input.path_forcing``
        and ``dir_input``. If not found uses the default path ``inmaps.nc`` in the
        root folder.

        If several files are used using '*' in ``input.path_forcing``, all corresponding
        files are read and merged into one xarray dataset before being split to one
        xarray DataArray per forcing variable in the hydromt ``forcing`` dictionary.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments to be passed to the `read_nc` method.
        """
        # Sort which path/ filename is actually the one used
        # Hierarchy is: 1: config, 2: default from component
        p = self.model.config.get_value("input.path_forcing") or self._filename
        # Check for input dir
        p_input = Path(self.model.config.get_value("dir_input", fallback=""), p)

        super().read(
            filename=p_input,
            mask_and_scale=False,
            **kwargs,
        )

    @hydromt_step
    def write(
        self,
        *,
        filename: str | None = None,
        output_frequency: str | None = None,
        time_chunk: int = 1,
        time_units="days since 1900-01-01T00:00:00",
        decimals: int = 2,
        overwrite: bool = False,
        **kwargs,
    ) -> None:
        """Write forcing model data.

        If no ``filename`` path is provided and path_forcing from the wflow toml exists,
        the following default filenames are used:

            * Default name format (with downscaling):
              inmaps_sourcePd_sourceTd_methodPET_freq_startyear_endyear.nc
            * Default name format (no downscaling):
              inmaps_sourceP_sourceT_methodPET_freq_startyear_endyear.nc

        Key-word arguments are passed to :py:meth:`~hydromt._io.writers.write_nc`

        Parameters
        ----------
        filename : str, optional
            Filepath to write the forcing data to, either absolute or relative to the
            model root. If None, the default filename will be used.
        output_frequency : str, optional
            The frequency of the files being written. If e.g. '3M' (3 months) is
            specified then files are written with a maximum of 3 months worth of data.
            By default None
        time_chunk: int, optional
            The chunk size for the time dimension when writing the forcing data.
            By default 1.
        time_units : str, optional
            Common time units when writing several netcdf forcing files.
            By default "days since 1900-01-01T00:00:00".
        decimals : int, optional
            Number of decimals to use when writing the forcing data.
            By default 2.
        overwrite : bool, optional
            Whether to overwrite existing files. Default is ``False`` unless the model
            is in w+ mode (FORCED_WRITE).
        **kwargs : dict
            Additional keyword arguments to be passed to the `write_nc` method.
        """
        self.root._assert_write_mode()
        overwrite = overwrite or self.root.mode.value == "w+"

        # Dont write if data is empty
        if self._data_is_empty():
            logger.warning(
                "Write forcing skipped: dataset is empty (no variables or data)."
            )
            return

        # Sort out the path
        # Hierarchy is: 1: signature, 2: config, 3: default
        filename = (
            filename
            or self.model.config.get_value("input.path_forcing")
            or self._filename
        )
        # Check for input dir
        p_input = Path(self.model.config.get_value("dir_input", fallback=""), filename)

        # Get the start and endtime
        starttime = self.model.config.get_value("time.starttime", fallback=None)
        endtime = self.model.config.get_value("time.endtime", fallback=None)

        # Logging the output
        logger.info("Write forcing file")
        start_time, end_time = self._validate_timespan(starttime, endtime)

        if Path(p_input).is_absolute():
            filepath = Path(p_input)
        else:
            filepath = (self.root.path / p_input).resolve()

        if filepath.exists() and not overwrite:
            logger.warning(
                f"Netcdf forcing file `{filepath}` already exists and overwriting "
                "is not enabled. To overwrite netcdf forcing file: change name "
                "`input.path_forcing` in setup_config section of the build "
                "inifile or allow overwriting with `overwrite` flag. A default "
                "name will be generated."
            )
            filepath = self._create_new_filename(
                filepath=filepath,
                start_time=start_time,
                end_time=end_time,
                frequency=output_frequency,
            )
            if filepath is None:  # should skip writing
                return

        # Clean-up forcing and write
        ds = self.data.drop_vars(["mask", "idx_out"], errors="ignore")

        if decimals is not None:
            ds = ds.round(decimals)

        # Write with output chunksizes with single timestep and complete
        # spatial grid to speed up the reading from wflow.jl
        # Dims are always ordered (time, y, x)
        ds.raster._check_dimensions()
        chunksizes = (
            time_chunk,
            ds.raster.ycoords.size,
            ds.raster.xcoords.size,
        )

        # Set the encoding for the outgoing dataset
        encoding = kwargs.pop("encoding", {})
        encoding.update(
            {
                v: {"dtype": "float32", "chunksizes": chunksizes}
                for v in ds.data_vars.keys()
            }
        )
        encoding["time"] = {"units": time_units}

        # Write the file either in one go
        if output_frequency is None:
            logger.info(f"Writing file {filepath.as_posix()}")
            close_handle = write_nc(
                ds,
                file_path=filepath,
                compress=True,
                gdal_compliant=True,
                rename_dims=True,
                force_sn=False,
                force_overwrite=True,
                progressbar=True,
                to_netcdf_kwargs={"encoding": encoding, **kwargs},
            )
            if close_handle is not None:
                self._deferred_file_close_handles.append(close_handle)
        # Or per period
        else:
            logger.info(f"Writing several forcing with freq {output_frequency}")
            # Updating path forcing in config
            for _, data_freq in ds.resample(time=output_frequency):
                # Sort out the outgoing filename
                start = data_freq["time"].dt.strftime("%Y%m%d")[0].item()
                filepath_freq = Path(filepath.parent, f"{filepath.stem}_{start}.nc")
                logger.info(f"Writing file {filepath_freq.as_posix()}")
                # Write to file
                close_handle = write_nc(
                    data_freq,
                    file_path=filepath_freq,
                    compress=True,
                    gdal_compliant=True,
                    rename_dims=True,
                    force_sn=False,
                    force_overwrite=True,
                    progressbar=True,
                    to_netcdf_kwargs={"encoding": encoding, **kwargs},
                )
                if close_handle is not None:
                    self._deferred_file_close_handles.append(close_handle)
            filepath = Path(filepath.parent, f"{filepath.stem}_*{filepath.suffix}")

        # Set back to the config
        self.model.config.set("input.path_forcing", filepath.as_posix())
        self.model.config.set("time.starttime", start_time.strftime(_DATETIME_FORMAT))
        self.model.config.set("time.endtime", end_time.strftime(_DATETIME_FORMAT))

    ## Mutating methods
    def set(
        self,
        data: xr.DataArray | xr.Dataset,
        name: str | None = None,
    ):
        """Add data to forcing.

        All layers in data must have identical spatial coordinates to existing
        forcing and staticmaps.

        Parameters
        ----------
        data: xarray.DataArray or xarray.Dataset
            new map layer to add to forcing
        name: str, optional
            Name of new map layer, this is used to overwrite the name of a DataArray
            and ignored if data is a Dataset
        """
        # Rename the temporal dimension to time
        if "time" not in data.dims:
            raise ValueError("'time' dimension not found in data")

        if self._region_component is not None:
            # Ensure the data is aligned with the region component (staticmaps)
            data = utils.align_grid_with_region(
                data,
                region_grid=self._get_grid_data(),
                region_component_name=self._region_component,
            )

            # Check for CRS - eg sediment forcing is wflow.jl output and has no crs
            if data.raster.crs is None:
                data.raster.set_crs(self._get_grid_data().raster.crs)

        # Call set of parent class
        super().set(
            data=data,
            name=name,
        )

    def clip(self):
        """Clip the forcing component to the region of the region component.

        If no region component is set, the forcing component is clipped to its own
        extent.

        """
        if self._region_component is None:
            logger.info(
                "No region component set, forcing component will not be clipped."
            )
            return

        if len(self.data) > 0:
            logger.info("Clipping forcing...")
            region_grid = self._get_grid_data()
            ds_forcing = self.data.raster.clip_bbox(region_grid.raster.bounds)
            self._data = xr.Dataset()  # clear existing forcing
            self.set(ds_forcing)

    ### Internal methods
    def _create_new_filename(
        self, filepath: Path, start_time: datetime, end_time: datetime, frequency: str
    ) -> Path:
        """Determine the filename to use for writing.

        Parameters
        ----------
        filepath : Path
            The filepath to use.

        Returns
        -------
        Path
            The determined filepath as a Path object.
        """
        if "*" in filepath.name:
            # get rid of * in case model had multiple forcing files and
            # write to single nc file.
            logger.warning("Writing multiple forcing files to one file")
            filepath = filepath.parent / filepath.name.replace("*", "")

        if not filepath.exists():
            return filepath

        # Create filename based on attributes
        sourceP = ""
        sourceT = ""
        methodPET = ""
        if "precip" in self.data:
            val = self.data["precip"].attrs.get("precip_clim_fn", None)
            Pdown = "d" if val is not None else ""
            val = self.data["precip"].attrs.get("precip_fn", None)
            if val is not None:
                sourceP = f"_{val}{Pdown}"
        if "temp" in self.data:
            val = self.data["temp"].attrs.get("temp_correction", "False")
            Tdown = "d" if val == "True" else ""
            val = self.data["temp"].attrs.get("temp_fn", None)
            if val is not None:
                sourceT = f"_{val}{Tdown}"
        if "pet" in self.data:
            val = self.data["pet"].attrs.get("pet_method", None)
            if val is not None:
                methodPET = f"_{val}"

        fn_default = f"inmaps{sourceP}{sourceT}{methodPET}_{frequency}_{start_time.year}_{end_time.year}.nc"  # noqa: E501
        filepath = filepath.parent / fn_default

        if not filepath.exists():
            return filepath

        logger.warning(
            f"Netcdf generated forcing file name `{filepath}` already exists, "
            "skipping write_forcing. "
        )
        return None

    def _validate_timespan(
        self, starttime: Optional[str] = None, endtime: Optional[str] = None
    ) -> tuple[datetime, datetime]:
        """Validate the timespan of the forcing data."""
        # Check if all dates between (starttime, endtime) are in all da forcing
        # Check if starttime and endtime timestamps are correct
        starttime = np.datetime64(
            starttime or np.datetime_as_string(min(self.data.time.values), unit="s")
        )
        if (
            starttime < min(self.data.time.values)
            or starttime not in self.data.time.values
        ):
            logger.warning(
                f"Start time {starttime} does not match the beginning of the data. "
                f"Changing to start of the data: {min(self.data.time.values)}."
            )
            starttime = min(self.data.time.values)
        starttime = pd.Timestamp(starttime).to_pydatetime()

        endtime = np.datetime64(
            endtime or np.datetime_as_string(max(self.data.time.values), unit="s")
        )
        if endtime > max(self.data.time.values) or endtime not in self.data.time.values:
            logger.warning(
                f"End time {endtime} does not match the end of the data. "
                f"Changing to end of the data: {max(self.data.time.values)}."
            )
            endtime = max(self.data.time.values)
        endtime = pd.Timestamp(endtime).to_pydatetime()

        if self._data_is_empty():
            logger.warning(
                "Write forcing skipped: dataset is empty (no variables or data). Ensure"
                " that forcing data is loaded before calling 'write'."
            )
            return (
                None,
                datetime.strptime(starttime, format=_DATETIME_FORMAT),
                datetime.strptime(endtime, format=_DATETIME_FORMAT),
            )

        return starttime, endtime

    def _data_is_empty(self) -> bool:
        """Check if the forcing data is empty."""
        return len(self.data.data_vars) == 0 or all(
            var.size == 0 for var in self.data.data_vars.values()
        )

    def test_equal(self, other: ModelComponent) -> tuple[bool, dict[str, str]]:
        """Test if two forcing components are equal.

        Checks the model component type as well as the data variables and their values.

        Parameters
        ----------
        other : ModelComponent
            The component to compare against.

        Returns
        -------
        tuple[bool, Dict[str, str]]
            True if the components are equal, and a dict with the associated errors per
            property checked.
        """
        errors: dict[str, str] = {}
        if not isinstance(other, self.__class__):
            errors["__class__"] = f"other does not inherit from {self.__class__}."

        # Check dimensions
        _, data_errors = utils.test_equal_grid_data(self.data, other.data)
        errors.update(data_errors)

        return len(errors) == 0, errors



"""Some utilities to upgrade Wflow model versions."""

import logging
from os.path import abspath, dirname, join
from pathlib import Path

import xarray as xr

from hydromt_wflow.components.tables import WflowTablesComponent
from hydromt_wflow.naming import (
    WFLOW_NAMES,
    WFLOW_SEDIMENT_NAMES,
    WFLOW_SEDIMENT_STATES_NAMES,
    WFLOW_STATES_NAMES,
)
from hydromt_wflow.utils import get_config, set_config
from hydromt_wflow.workflows.reservoirs import (
    RESERVOIR_COMMON_PARAMETERS,
    RESERVOIR_CONTROL_PARAMETERS,
    RESERVOIR_LAYERS_SEDIMENT,
    RESERVOIR_UNCONTROL_PARAMETERS,
    merge_reservoirs,
    merge_reservoirs_sediment,
    set_rating_curve_layer_data_type,
)

logger = logging.getLogger(f"hydromt.{__name__}")

DATADIR = Path(join(dirname(abspath(__file__)), "data"))

__all__ = [
    "convert_to_wflow_v1_sbm",
    "convert_to_wflow_v1_sediment",
]


def _solve_var_name(var: str | dict, path: str, add: list):
    """Solve the config file into individual entries.

    Every entry is the entire path ("river.lateral.< something >") plus its value.

    Parameters
    ----------
    var : str | dict,
        Either the direct settings entry or a dictionary containing nested settings.
    path : str,
        Prepend the entries with the value (e.g. "lateral" or "lateral.river")
    add : list
        Usually an empty list in which the temporary headers are stored.
    """
    if not isinstance(var, dict):
        sep = "." if path else ""
        add_str = ".".join(add) if add else ""
        yield (var, path + sep + add_str)
        return
    for key, item in var.items():
        yield from _solve_var_name(item, path, add + [key])


def _create_v0_to_v1_var_mapping(wflow_vars: dict) -> dict:
    output_dict = {}
    for value in wflow_vars.values():
        if isinstance(value["wflow_v0"], list):
            for var in value["wflow_v0"]:
                output_dict[var] = value["wflow_v1"]
        else:
            output_dict[value["wflow_v0"]] = value["wflow_v1"]
    return output_dict


def _set_input_vars(
    wflow_v0_var: str,
    wflow_v1_var: str | None,
    config_in: dict,
    config_out: dict,
    input_options: dict,
    input_variables: dict,
    forcing_variables: list[str],
    cyclic_variables: list[str],
) -> dict:
    name = get_config(key=f"input.{wflow_v0_var}", config=config_in, fallback=None)
    if name is not None and wflow_v1_var is not None:
        if isinstance(name, dict) and "netcdf" in name.keys():
            name["netcdf_variable_name"] = name["netcdf"]["variable"]["name"]
            del name["netcdf"]
        if wflow_v0_var in input_options.keys():
            return config_out
        elif wflow_v0_var in input_variables:
            config_out["input"][wflow_v1_var] = name
        elif wflow_v0_var in forcing_variables:
            config_out["input"]["forcing"][wflow_v1_var] = name
        elif wflow_v0_var in cyclic_variables:
            config_out["input"]["cyclic"][wflow_v1_var] = name
        else:
            config_out["input"]["static"][wflow_v1_var] = name
    return config_out


def _convert_to_wflow_v1(
    config: dict,
    wflow_vars: dict,
    states_vars: dict,
    model_options: dict = {},
    cross_options: dict = {},
    input_options: dict = {},
    input_variables: list = [],
    additional_variables: dict = {},
) -> dict:
    """Convert the config to Wflow v1 format.

    Parameters
    ----------
    config: dict
        The config to convert.
    wflow_vars: dict
        The Wflow variables dict to use for the conversion between versions.
        Either WFLOW_NAMES or WFLOW_SEDIMENT_NAMES.
    states_vars: dict
        The Wflow states variables dict to use for the conversion between versions.
        Either WFLOW_STATES_NAMES or WFLOW_SEDIMENT_STATES_NAMES.
    model_options: dict, optional
        Options in the [model] section of the TOML that were updated in Wflow v1.
    input_options: dict, optional
        Options in the [input] section of the TOML that were updated in Wflow v1.
    input_variables: list, optional
        Variables that were moved to input rather than input.static.

    Returns
    -------
    config_out: dict
        The converted config.
    """
    if config["model"].get("reinit") == False:
        logger.warning(
            "Converting states is not supported by this conversion code, therefore the "
            "reinit option (new name cold_start__flag) is adjusted to True. If model "
            "states are required, please produce new states using a Wflow simulation or"
            " use the setup_cold_states function."
        )
        config["model"]["reinit"] = True
    WFLOW_CONVERSION = _create_v0_to_v1_var_mapping(wflow_vars)
    for k, v in states_vars.items():
        WFLOW_CONVERSION[v["wflow_v0"]] = v["wflow_v1"]
    # Add a few extra output variables that are supported by the conversion
    WFLOW_CONVERSION.update(additional_variables)

    # Logging function for the other non supported variables
    def _warn_str(wflow_var, output_type):
        logger.warning(
            f"Output variable {wflow_var} not supported for the conversion. "
            f"Skipping from {output_type} output."
        )

    # Update function for the output.netcdf_grid
    def _update_output_netcdf_grid(wflow_var, var_name):
        if wflow_var in WFLOW_CONVERSION.keys():
            config_out["output"]["netcdf_grid"]["variables"][
                WFLOW_CONVERSION[wflow_var]
            ] = var_name
        else:
            _warn_str(var_name, "netcdf_grid")

    # Initialize the output config
    logger.info("Converting config to Wflow v1 format")
    logger.info("Converting config general, time and model sections")
    config_out = {}

    # Start with the general section - split into general, time and logging in v1
    input_section = {
        "general": ["dir_input", "dir_output", "fews_run"],
        "time": [
            "calendar",
            "starttime",
            "endtime",
            "time_units",
            "timestepsecs",
        ],
        "logging": ["loglevel", "path_log", "silent"],
    }
    for section, options in input_section.items():
        for key in options:
            value = get_config(key=key, config=config, fallback=None)
            if value is not None:
                if section == "general":
                    # only fews_run was renamed in general options
                    if key == "fews_run":
                        key = "fews_run__flag"
                    config_out[key] = value
                else:
                    config_out[section] = config_out.get(section, {})
                    config_out[section][key] = value

    # Model section
    config_out["model"] = {}
    for value, config_var in _solve_var_name(config["model"], "", []):
        if config_var not in model_options:
            # Model option that did not change in v1
            set_config(config_out, f"model.{config_var}", value)
        else:
            new_config_var = model_options[config_var]
            if isinstance(new_config_var, (list, tuple)):
                for elem in new_config_var:
                    set_config(config_out, f"model.{elem}", value)
                continue
            elif new_config_var is None:
                # This option was moved to cross_options
                continue
            set_config(config_out, f"model.{new_config_var}", value)

    # Routing options were renamed
    routing_rename = {
        "kinematic-wave": "kinematic_wave",
        "local-inertial": "local_inertial",
    }
    for key in ["river_routing", "land_routing"]:
        routing_option = get_config(key=f"model.{key}", config=config, fallback=None)
        if routing_option is not None:
            set_config(config_out, f"model.{key}", routing_rename.get(routing_option))

    # State
    logger.info("Converting config state section")
    config_out["state"] = {
        "path_input": get_config(
            key="state.path_input", config=config, fallback="instate/instates.nc"
        ),
        "path_output": get_config(
            key="state.path_output", config=config, fallback="outstate/outstates.nc"
        ),
    }
    # Go through the states variables
    config_out["state"]["variables"] = {}
    for key, variables in states_vars.items():
        name = get_config(
            key=f"state.{variables['wflow_v0']}", config=config, fallback=None
        )
        if name is not None and variables["wflow_v1"] is not None:
            config_out["state"]["variables"][variables["wflow_v1"]] = name

    # Input section
    logger.info("Converting config input section")
    cyclic_variables = []
    forcing_variables = []

    reservoir_locs = []
    if get_config(config, "input.lateral.river.reservoir.locs") is not None:
        reservoir_locs.append(get_config(config, "input.lateral.river.reservoir.locs"))
    if get_config(config, "input.lateral.river.lake.locs") is not None:
        reservoir_locs.append(get_config(config, "input.lateral.river.lake.locs"))
    reservoir_locs_map = []

    config_out["input"] = {}
    for key, name in config["input"].items():
        if key in input_options.keys():
            config_out["input"][input_options[key]] = name
        elif key == "forcing":
            forcing_variables = name
        elif key == "cyclic":
            cyclic_variables = name
        elif key not in ["vertical", "lateral"]:  # variables are done separately
            # For reservoir / lake locations map, skip
            if name in reservoir_locs:
                reservoir_locs_map.append(key)
            else:
                config_out["input"][key] = name

    # Go through the input variables
    config_out["input"]["forcing"] = {}
    config_out["input"]["cyclic"] = {}
    config_out["input"]["static"] = {}
    for key, variables in wflow_vars.items():
        if isinstance(variables["wflow_v0"], list):
            for wflow_v0_var in variables["wflow_v0"]:
                config_out = _set_input_vars(
                    wflow_v0_var=wflow_v0_var,
                    wflow_v1_var=variables["wflow_v1"],
                    config_in=config,
                    config_out=config_out,
                    input_options=input_options,
                    input_variables=input_variables,
                    forcing_variables=forcing_variables,
                    cyclic_variables=cyclic_variables,
                )
        else:
            config_out = _set_input_vars(
                wflow_v0_var=variables["wflow_v0"],
                wflow_v1_var=variables["wflow_v1"],
                config_in=config,
                config_out=config_out,
                input_options=input_options,
                input_variables=input_variables,
                forcing_variables=forcing_variables,
                cyclic_variables=cyclic_variables,
            )

    # Cross options
    for opt_old, opt_new in cross_options.items():
        value = get_config(key=opt_old, config=config)
        if value is None:
            continue
        # Ensure that it is set either as a value or as a map
        elif isinstance(value, str):
            set_config(config_out, opt_new, value)
        else:
            set_config(config_out, f"{opt_new}.value", value)

    # Output netcdf_grid section
    logger.info("Converting config output sections")
    if get_config(key="output", config=config, fallback=None) is not None:
        config_out["output"] = {}
        config_out["output"]["netcdf_grid"] = {
            "path": get_config(key="output.path", config=config, fallback="output.nc"),
            "compressionlevel": get_config(
                key="output.compressionlevel", config=config, fallback=1
            ),
        }
        config_out["output"]["netcdf_grid"]["variables"] = {}
        for key, value in config["output"].items():
            if key in ["path", "compressionlevel"]:
                continue

            for var_name, wflow_var in _solve_var_name(value, key, []):
                _update_output_netcdf_grid(wflow_var, var_name)

    # Output netcdf_scalar section
    map_variable_conversions = {
        "lateral.river.lake.locs": "reservoir_location__count",
        "lateral.river.reservoir.locs": "reservoir_location__count",
    }
    # Add the reservoir maps that were potentially in the input section
    for var in reservoir_locs_map:
        map_variable_conversions[var] = "reservoir_location__count"

    if get_config(key="netcdf", config=config, fallback=None) is not None:
        if "output" not in config_out:
            config_out["output"] = {}
        config_out["output"]["netcdf_scalar"] = {
            "path": get_config(
                key="netcdf.path", config=config, fallback="output_scalar.nc"
            ),
        }
        config_out["output"]["netcdf_scalar"]["variable"] = []
        nc_scalar_vars = get_config(key="netcdf.variable", config=config, fallback=[])
        for nc_scalar in nc_scalar_vars:
            if nc_scalar["parameter"] in WFLOW_CONVERSION:
                nc_scalar["parameter"] = WFLOW_CONVERSION[nc_scalar["parameter"]]
                if map_var := nc_scalar.get("map"):
                    if map_var in input_options:
                        nc_scalar["map"] = input_options[map_var]
                    elif map_var in map_variable_conversions:
                        nc_scalar["map"] = map_variable_conversions[map_var]
                config_out["output"]["netcdf_scalar"]["variable"].append(nc_scalar)
            else:
                _warn_str(nc_scalar["parameter"], "netcdf_scalar")

    # Output csv section
    if get_config(key="csv", config=config, fallback=None) is not None:
        if "output" not in config_out:
            config_out["output"] = {}
        config_out["output"]["csv"] = {}

        config_out["output"]["csv"]["path"] = get_config(
            key="csv.path", config=config, fallback="output.csv"
        )
        config_out["output"]["csv"]["column"] = []
        csv_vars = get_config(key="csv.column", config=config, fallback=[])
        for csv_var in csv_vars:
            if csv_var["parameter"] in WFLOW_CONVERSION:
                csv_var["parameter"] = WFLOW_CONVERSION[csv_var["parameter"]]
                if map_var := csv_var.get("map"):
                    if map_var in input_options:
                        csv_var["map"] = input_options[map_var]
                    elif map_var in map_variable_conversions:
                        csv_var["map"] = map_variable_conversions[map_var]
                config_out["output"]["csv"]["column"].append(csv_var)
            else:
                _warn_str(csv_var["parameter"], "csv")

    return config_out


def convert_to_wflow_v1_sbm(config: dict) -> dict:
    """Convert the config to Wflow v1 format for SBM.

    Parameters
    ----------
    config: dict
        The config to convert.

    Returns
    -------
    config_out: dict
        The converted config.
    """
    additional_variables = {
        "vertical.interception": "vegetation_canopy_water__interception_volume_flux",
        "vertical.actevap": "land_surface__evapotranspiration_volume_flux",
        "vertical.actinfilt": "soil_water__infiltration_volume_flux",
        "vertical.excesswatersoil": "compacted_soil_surface_water__excess_volume_flux",
        "vertical.excesswaterpath": "non_compacted_soil_surface_water__excess_volume_flux",  # noqa : E501
        "vertical.exfiltustore": "soil_surface_water_unsaturated_zone__exfiltration_volume_flux",  # noqa : E501
        "vertical.exfiltsatwater": "land.soil.variables.exfiltsatwater",
        "vertical.recharge": "soil_water_saturated_zone_top__net_recharge_volume_flux",
        "vertical.vwc_percroot": "soil_water_root_zone__volume_percentage",
        "lateral.land.q_av": "land_surface_water__volume_flow_rate",
        "lateral.land.h_av": "land_surface_water__depth",
        "lateral.land.to_river": "land_surface_water__to_river_volume_flow_rate",
        "lateral.subsurface.to_river": "subsurface_water__to_river_volume_flow_rate",
        "lateral.subsurface.drain.flux": "land_drain_water__to_subsurface_volume_flow_rate",  # noqa : E501
        "lateral.subsurface.flow.aquifer.head": "subsurface_water__hydraulic_head",
        "lateral.subsurface.river.flux": "river_water__to_subsurface_volume_flow_rate",
        "lateral.subsurface.recharge.rate": "subsurface_water_saturated_zone_top__net_recharge_volume_flow_rate",  # noqa : E501
        "lateral.river.q_av": "river_water__volume_flow_rate",
        "lateral.river.h_av": "river_water__depth",
        "lateral.river.volume": "river_water__volume",
        "lateral.river.inwater": "river_water__lateral_inflow_volume_flow_rate",
        "lateral.river.floodplain.volume": "floodplain_water__volume",
        "lateral.river.reservoir.volume": "reservoir_water__volume",
        "lateral.river.reservoir.totaloutflow": "reservoir_water__outgoing_volume_flow_rate",  # noqa : E501
        "lateral.river.reservoir.outflow": "reservoir_water__outgoing_volume_flow_rate",
        "lateral.river.reservoir.inflow": "reservoir_water__incoming_volume_flow_rate",
        "lateral.river.reservoir.precipitation": "reservoir_water__precipitation_volume_flux",  # noqa : E501
        "lateral.river.reservoir.evaporation": "reservoir_water__potential_evaporation_volume_flux",  # noqa : E501
        "lateral.river.reservoir.actevap": "reservoir_water__evaporation_volume_flux",
        "lateral.river.lake.storage": "reservoir_water__volume",
        "lateral.river.lake.totaloutflow": "reservoir_water__outgoing_volume_flow_rate",
        "lateral.river.lake.outflow": "reservoir_water__outgoing_volume_flow_rate",
        "lateral.river.lake.inflow": "reservoir_water__incoming_volume_flow_rate",
        "lateral.river.lake.waterlevel": "reservoir_water_surface__elevation",
        "lateral.river.lake.precipitation": "reservoir_water__precipitation_volume_flux",  # noqa : E501
        "lateral.river.lake.evaporation": "reservoir_water__potential_evaporation_volume_flux",  # noqa : E501
        "lateral.river.lake.actevap": "reservoir_water__evaporation_volume_flux",
    }

    # Options in model section that were renamed
    model_options = {
        "reinit": "cold_start__flag",
        "sizeinmetres": "cell_length_in_meter__flag",
        "reservoirs": "reservoir__flag",
        "lakes": "reservoir__flag",
        "snow": "snow__flag",
        "glacier": "glacier__flag",
        "pits": "pit__flag",
        "masswasting": "snow_gravitational_transport__flag",
        "thicknesslayers": "soil_layer__thickness",
        "min_streamorder_land": "land_streamorder__min_count",
        "min_streamorder_river": "river_streamorder__min_count",
        "drains": "drain__flag",
        "kin_wave_iteration": "kinematic_wave__adaptive_time_step_flag",
        "kw_land_tstep": "land_kinematic_wave__time_step",
        "kw_river_tstep": "river_kinematic_wave__time_step",
        "inertial_flow_alpha": [
            "river_local_inertial_flow__alpha_coefficient",
            "land_local_inertial_flow__alpha_coefficient",
        ],
        "h_thresh": [
            "river_water_flow_threshold__depth",
            "land_surface_water_flow_threshold__depth",
        ],
        "froude_limit": [
            "river_water_flow__froude_limit_flag",
            "land_surface_water_flow__froude_limit_flag",
        ],
        "floodplain_1d": "floodplain_1d__flag",
        "inertial_flow_theta": "land_local_inertial_flow__theta_coefficient",
        "soilinfreduction": "soil_infiltration_reduction__flag",
        "transfermethod": "topog_sbm_transfer__flag",
        "water_demand.domestic": "water_demand.domestic__flag",
        "water_demand.industry": "water_demand.industry__flag",
        "water_demand.livestock": "water_demand.livestock__flag",
        "water_demand.paddy": "water_demand.paddy__flag",
        "water_demand.nonpaddy": "water_demand.nonpaddy__flag",
        "constanthead": "constanthead__flag",
        "riverlength_bc": None,  # moved to cross_options
    }

    # Options in input section that were renamed
    input_options = {
        "ldd": "basin__local_drain_direction",
        "river_location": "river_location__mask",
        "subcatchment": "subbasin_location__count",
    }

    # variables that were moved to input rather than input.static
    input_variables = [
        "lateral.river.lake.areas",
        "lateral.river.lake.locs",
        "lateral.river.lake.linkedlakelocs",
        "lateral.river.reservoir.areas",
        "lateral.river.reservoir.locs",
    ]

    # Wflow entries that cross main headers (i.e. [input, state, model, output])
    cross_options = {
        "input.lateral.subsurface.conductivity_profile": "model.conductivity_profile",
        "model.riverlength_bc": "input.static.model_boundary_condition_river__length",
    }

    config_out = _convert_to_wflow_v1(
        config=config,
        wflow_vars=WFLOW_NAMES,
        states_vars=WFLOW_STATES_NAMES,
        model_options=model_options,
        cross_options=cross_options,
        input_options=input_options,
        input_variables=input_variables,
        additional_variables=additional_variables,
    )

    return config_out


def convert_to_wflow_v1_sediment(config: dict) -> dict:
    """Convert the config to Wflow v1 format for sediment.

    Parameters
    ----------
    config: dict
        The config to convert.

    Returns
    -------
    config_out: dict
        The converted config.
    """
    additional_variables = {
        "vertical.soilloss": "soil_erosion__mass_flow_rate",
        "lateral.river.SSconc": "river_water_sediment__suspended_mass_concentration",
        "lateral.river.Bedconc": "river_water_sediment__bedload_mass_concentration",
        "lateral.river.outsed": "land_surface_water_sediment__mass_flow_rate",
        "lateral.land.inlandclay": "land_surface_water_clay__to_river_mass_flow_rate",
        "lateral.land.inlandsilt": "land_surface_water_silt__to_river_mass_flow_rate",
        "lateral.land.inlandsand": "land_surface_water_sand__to_river_mass_flow_rate",
        "lateral.land.inlandsagg": "land_surface_water_small_aggregates__to_river_mass_flow_rate",  # noqa: E501
        "lateral.land.inlandlagg": "land_surface_water_large_aggregates__to_river_mass_flow_rate",  # noqa: E501
    }

    # Options in model section that were renamed
    model_options = {
        "reinit": "cold_start__flag",
        "sizeinmetres": "cell_length_in_meter__flag",
        "runrivermodel": "run_river_model__flag",
        "doreservoir": "reservoir__flag",
        "dolake": "reservoir__flag",
        "rainerosmethod": "rainfall_erosion",
        "landtransportmethod": "land_transport",
        "rivtransportmethod": "river_transport",
    }

    # Options in input section that were renamed
    input_options = {
        "ldd": "basin__local_drain_direction",
        "river_location": "river_location__mask",
        "subcatchment": "subbasin_location__count",
    }

    # variables that were moved to input rather than input.static
    input_variables = [
        "vertical.lakeareas",
        "lateral.river.lakelocs",
        "vertical.resareas",
        "lateral.river.reslocs",
    ]

    config_out = _convert_to_wflow_v1(
        config=config,
        wflow_vars=WFLOW_SEDIMENT_NAMES,
        states_vars=WFLOW_SEDIMENT_STATES_NAMES,
        model_options=model_options,
        input_options=input_options,
        input_variables=input_variables,
        additional_variables=additional_variables,
    )

    return config_out


def convert_reservoirs_to_wflow_v1_sbm(
    grid: xr.Dataset,
    config: dict,
) -> tuple[xr.Dataset | None, list[str], dict[str, str]]:
    """
    Merge reservoirs and lakes layers from a v0.x model config.

    Returns None if no reservoirs or lakes are present in the config.
    If reservoirs are present, also add rating_curve, storage_curve and initial_depth.
    The output variables will use standard names for the variables (e.g.
    "reservoir_area_id").

    Parameters
    ----------
    grid : xr.Dataset
        The grid dataset containing the reservoirs and lakes layers.
    config : dict
        The model configuration dictionary.

    Returns
    -------
    xr.Dataset | None
        The merged dataset with reservoirs and lakes layers, or None if not applicable.
    list[str]
        List of variables to remove from the grid (will be replaced by the merged ones).
    dict[str, str]
        Dictionary of wflow v1 config options to update with the new variable names.
    """
    variables_to_remove = []
    config_options = {}
    ds_res = xr.Dataset()

    has_reservoirs = get_config(key="model.reservoirs", config=config, fallback=False)
    has_lakes = get_config(key="model.lakes", config=config, fallback=False)

    if not has_reservoirs and not has_lakes:
        logger.info("No reservoirs or lakes found in the config. Skipping conversion.")
        return None, variables_to_remove, config_options
    else:
        logger.info(
            "Reservoirs/lakes found in the grid, converting to Wflow v1 format."
        )
        # config options will need to be updated with the standard names
        config_options["model.reservoir__flag"] = True
        config_options["state.variables.reservoir_water_surface__elevation"] = (
            "reservoir_water_level"
        )

    # Start with the reservoir layers
    if has_reservoirs:
        reservoir_layers = (
            RESERVOIR_COMMON_PARAMETERS
            + RESERVOIR_CONTROL_PARAMETERS
            + [
                "reservoir_area_id",
                "reservoir_outlet_id",
            ]
        )
        for layer in reservoir_layers:
            if layer in [
                "reservoir_initial_depth",
                "reservoir_rating_curve",
                "reservoir_storage_curve",
            ]:
                # These layers are not in the config, so we skip them
                continue
            wflow_var = WFLOW_NAMES[layer]["wflow_v0"]
            # get the map name from config
            map_name = get_config(key=f"input.{wflow_var}", config=config)
            # add to the variables to remove
            variables_to_remove.append(map_name)
            if map_name in grid:
                ds_res[layer] = grid[map_name].copy()
            else:
                logger.warning(f"Reservoir layer {map_name} not found in the grid.")

        # Add the additional layers that are not in the config
        res_mask_inv = (
            ds_res["reservoir_outlet_id"] == ds_res["reservoir_outlet_id"].raster.nodata
        )
        ds_res["reservoir_rating_curve"] = ds_res["reservoir_outlet_id"].where(
            res_mask_inv,
            4,
        )
        ds_res["reservoir_storage_curve"] = ds_res["reservoir_outlet_id"].where(
            res_mask_inv,
            1,
        )
        ds_res["reservoir_initial_depth"] = (
            ds_res["reservoir_target_full_fraction"]
            * ds_res["reservoir_max_volume"]
            / ds_res["reservoir_area"]
        )

        # Update the config options
        for layer in reservoir_layers:
            # layers that are in "main" input section
            if layer in [
                "reservoir_area_id",
                "reservoir_outlet_id",
                "reservoir_lower_id",
            ]:
                wflow_var_v1 = f"input.{WFLOW_NAMES[layer]['wflow_v1']}"
            else:
                # Find if in cyclic / forcing / static
                v0_var = WFLOW_NAMES[layer]["wflow_v0"]
                if v0_var in get_config(config=config, key="input.cyclic", fallback=[]):
                    wflow_var_v1 = f"input.cyclic.{WFLOW_NAMES[layer]['wflow_v1']}"
                elif v0_var in get_config(
                    config=config, key="input.forcing", fallback=[]
                ):
                    wflow_var_v1 = f"input.forcing.{WFLOW_NAMES[layer]['wflow_v1']}"
                else:
                    wflow_var_v1 = f"input.static.{WFLOW_NAMES[layer]['wflow_v1']}"
            config_options[wflow_var_v1] = layer

    # Move to the lake layers
    if has_lakes:
        lake_layers = RESERVOIR_UNCONTROL_PARAMETERS + [
            "lake_area_id",
            "lake_outlet_id",
            "lake_area",
            "reservoir_initial_depth",
            "reservoir_rating_curve",
            "reservoir_storage_curve",
        ]
        ds_lakes = xr.Dataset()

        for layer in lake_layers:
            wflow_var = WFLOW_NAMES[layer]["wflow_v0"]
            # get the map name from config
            map_name = get_config(key=f"input.{wflow_var}", config=config)
            # add to the variables to remove
            variables_to_remove.append(map_name)
            if map_name in grid:
                # Replace lake by reservoir in the output layer name
                layer_out = layer.replace("lake", "reservoir")
                ds_lakes[layer_out] = grid[map_name].copy()

        # Merge lakes into the reservoirs dataset if needed
        if not has_reservoirs:
            ds_res = ds_lakes
        else:
            ds_merge = merge_reservoirs(
                ds_lakes,
                ds_res,
                duplicate_id="skip",
            )
            if ds_merge is None:
                logger.warning(
                    "Merging lakes into reservoirs failed because of duplicated IDs. "
                    "Only reservoirs will be converted and you may need to update the "
                    "models to add the missing lakes manually."
                )
            else:
                ds_res = ds_merge

        # Update the config options
        for layer in lake_layers:
            layer_out = layer.replace("lake", "reservoir")
            if layer in ["lake_area_id", "lake_outlet_id", "reservoir_lower_id"]:
                wflow_var_v1 = f"input.{WFLOW_NAMES[layer_out]['wflow_v1']}"
            else:
                wflow_var_v1 = f"input.static.{WFLOW_NAMES[layer_out]['wflow_v1']}"
            config_options[wflow_var_v1] = layer_out

    # Ensure correct data types for rating curve
    ds_res = set_rating_curve_layer_data_type(ds_res)
    return ds_res, variables_to_remove, config_options


def convert_reservoirs_to_wflow_v1_sediment(
    grid: xr.Dataset,
    config: dict,
) -> tuple[xr.Dataset | None, list[str], dict[str, str]]:
    """
    Merge reservoirs and lakes layers from a v0.x model config.

    Returns None if no reservoirs or lakes are present in the config.
    If lakes are present, also add trapping_efficiency of 0. The output variables will
    use standard names for the variables (e.g. "reservoir_area_id").

    Parameters
    ----------
    grid : xr.Dataset
        The grid dataset containing the reservoirs and lakes layers.
    config : dict
        The model configuration dictionary.

    Returns
    -------
    xr.Dataset | None
        The merged dataset with reservoirs and lakes layers, or None if not applicable.
    list[str]
        List of variables to remove from the grid (will be replaced by the merged ones).
    dict[str, str]
        Dictionary of wflow v1 config options to update with the new variable names.
    """
    variables_to_remove = []
    config_options = {}
    ds_res = xr.Dataset()

    has_reservoirs = get_config(key="model.doreservoir", config=config, fallback=False)
    has_lakes = get_config(key="model.dolake", config=config, fallback=False)

    if not has_reservoirs and not has_lakes:
        logger.info("No reservoirs or lakes found in the config. Skipping conversion.")
        return None, variables_to_remove, config_options
    else:
        logger.info(
            "Reservoirs/lakes found in the grid, converting to Wflow v1 format."
        )
        # config options will need to be updated with the standard names
        config_options["model.reservoir__flag"] = True
        for layer in RESERVOIR_LAYERS_SEDIMENT:
            if layer in ["reservoir_area_id", "reservoir_outlet_id"]:
                wflow_var_v1 = f"input.{WFLOW_SEDIMENT_NAMES[layer]['wflow_v1']}"
            else:
                wflow_var_v1 = f"input.static.{WFLOW_SEDIMENT_NAMES[layer]['wflow_v1']}"
            config_options[wflow_var_v1] = layer

    # Start with the reservoir layers
    if has_reservoirs:
        for layer in RESERVOIR_LAYERS_SEDIMENT:
            wflow_var = WFLOW_SEDIMENT_NAMES[layer]["wflow_v0"]
            # get the map name from config
            map_name = get_config(key=f"input.{wflow_var}", config=config)
            # add to the variables to remove
            variables_to_remove.append(map_name)
            if map_name in grid:
                ds_res[layer] = grid[map_name].copy()
            else:
                logger.warning(f"Reservoir layer {map_name} not found in the grid.")

    # Move to the lake layers
    if has_lakes:
        ds_lakes = xr.Dataset()

        lake_layers = [
            "lake_area_id",
            "lake_outlet_id",
            "lake_area",
            "reservoir_trapping_efficiency",
        ]
        for layer in lake_layers:
            if layer == "reservoir_trapping_efficiency":
                # This layer is not in the config, so we skip it
                continue
            wflow_var = WFLOW_SEDIMENT_NAMES[layer]["wflow_v0"]
            # get the map name from config
            map_name = get_config(key=f"input.{wflow_var}", config=config)
            # add to the variables to remove
            variables_to_remove.append(map_name)
            if map_name in grid:
                # Replace lake by reservoir in the output layer name
                layer_out = layer.replace("lake", "reservoir")
                ds_lakes[layer_out] = grid[map_name].copy()

        # Add trapping efficiency
        ds_lakes["reservoir_trapping_efficiency"] = ds_lakes[
            "reservoir_outlet_id"
        ].where(
            ds_lakes["reservoir_outlet_id"]
            == ds_lakes["reservoir_outlet_id"].raster.nodata,
            0.0,
        )

        # Merge lakes into the reservoirs dataset if needed
        if not has_reservoirs:
            ds_res = ds_lakes
        else:
            ds_merge = merge_reservoirs_sediment(
                ds_lakes,
                ds_res,
                duplicate_id="skip",
            )
            if ds_merge is None:
                logger.warning(
                    "Merging lakes into reservoirs failed because of duplicated IDs. "
                    "Only reservoirs will be converted and you may need to update the "
                    "models to add the missing lakes manually."
                )
            else:
                ds_res = ds_merge

    return ds_res, variables_to_remove, config_options


def upgrade_lake_tables_to_reservoir_tables_v1(tables: WflowTablesComponent) -> None:
    logger.info("Renaming lake_*.csv files to reservoir_*.csv files.")
    for key in [t for t in tables.data.keys() if t.startswith("lake_")]:
        data_table = tables.data.pop(key)
        new_name = key.replace("lake_", "reservoir_")
        tables.set(data_table, name=new_name)
        logger.debug(f"Renamed table {key} to {new_name}.")



import shutil
from collections.abc import Callable
from pathlib import Path
from unittest.mock import MagicMock

import numpy as np
import pytest
import xarray as xr

from hydromt_wflow.components import WflowStatesComponent
from hydromt_wflow.wflow_sbm import WflowSbmModel


@pytest.fixture
def mock_model(mock_model_factory) -> WflowSbmModel:
    """Fixture to create a mock WflowSbmModel."""
    return mock_model_factory(mode="w")


def test_wflow_states_component_init(mock_model: WflowSbmModel):
    # Setup the component
    component = WflowStatesComponent(mock_model)

    # Assert that the internal data is None
    assert component._data is None

    # When asking for data property, it should return a xarray.Dataset
    assert isinstance(component.data, xr.Dataset)
    assert isinstance(component._data, xr.Dataset)  # Same for internal
    assert len(component.data) == 0


def test_wflow_states_component_init_with_region(mock_model_staticmaps: WflowSbmModel):
    # Setup the component with a region component
    component = WflowStatesComponent(
        mock_model_staticmaps, region_component="staticmaps"
    )

    # Check if the region component is set correctly
    assert component._region_component == "staticmaps"
    assert "basin" in component._get_grid_data()


def test_wflow_states_component_set(
    mock_model_staticmaps: WflowSbmModel, grid_dummy_data: xr.DataArray
):
    # Setup the component
    component = WflowStatesComponent(
        mock_model_staticmaps, region_component="staticmaps"
    )

    # Assert that the internal data is None, initializing will happen in set
    assert component._data is None

    # Set an entry
    # Copy else the name gets altered for the next test?
    component.set(grid_dummy_data.copy(), name="test_layer")

    # Assert the content
    assert isinstance(component._data, xr.Dataset)
    assert "test_layer" in component.data
    assert len(component.data) == 1


def test_wflow_states_component_set_alt(
    mock_model_staticmaps: WflowSbmModel, grid_dummy_data: xr.DataArray
):
    # Setup the component
    component = WflowStatesComponent(
        mock_model_staticmaps, region_component="staticmaps"
    )

    # Try with dataset
    component.set(grid_dummy_data.to_dataset())

    # Assert the content
    assert isinstance(component._data, xr.Dataset)
    assert "dummy_grid" in component.data


def test_wflow_states_component_set_errors(
    mock_model_staticmaps: WflowSbmModel, grid_dummy_data: xr.DataArray
):
    # Setup the component
    component = WflowStatesComponent(
        mock_model_staticmaps, region_component="staticmaps"
    )

    # Try with np.ndarray
    array = grid_dummy_data.values
    with pytest.raises(TypeError, match="Data must be an xarray Dataset or DataArray"):
        component.set(array)

    # Try with a different grid that is not identical
    # This should raise an error
    different_grid = grid_dummy_data.copy()
    different_grid["x"] = different_grid.x + 1  # Shift x coordinates
    with pytest.raises(
        ValueError, match="Data grid must be identical to staticmaps component"
    ):
        component.set(different_grid)


def test_wflow_states_component_read(
    mock_model_factory: Callable[[Path, str], WflowSbmModel],
    model_subbasin_cached: Path,
):
    # Set it to read mode
    mock_model = mock_model_factory(path=model_subbasin_cached, mode="r")

    # Setup the component
    component = WflowStatesComponent(mock_model)

    # Assert its data currently none
    assert component._data is None

    # Read the data
    type(component.model.config).get_value = MagicMock(return_value="")
    component.read()

    # Assert the read data
    assert isinstance(component.data, xr.Dataset)
    assert len(component.data) == 13
    assert "river_instantaneous_q" in component.data


def test_wflow_states_component_read_init(
    mock_model_factory: Callable[[Path, str], WflowSbmModel],
    model_subbasin_cached: Path,
):
    # Set it to read mode
    mock_model = mock_model_factory(path=model_subbasin_cached, mode="r")

    # Setup the component
    component = WflowStatesComponent(mock_model)
    type(component.model.config).get_value = MagicMock(return_value="")
    assert component._data is None  # Assert no data or structure yet

    # Read at init
    assert len(component.data) == 13
    assert "river_instantaneous_q" in component.data


def test_wflow_states_component_read_no_crs(
    mock_model_factory: Callable[[Path, str], WflowSbmModel],
    tmp_path: Path,
    model_subbasin_cached: Path,
    caplog,
):
    # Create a copy of the model files to modify
    model_copy_path = tmp_path / "model_copy"
    model_copy_path.mkdir(exist_ok=True)

    shutil.copytree(model_subbasin_cached, model_copy_path, dirs_exist_ok=True)

    # Remove CRS from the states file
    states_file = model_subbasin_cached / "instate" / "instates.nc"
    ds = xr.open_dataset(states_file)
    ds.raster.set_crs(None)  # Remove CRS
    ds.to_netcdf(model_copy_path / "instate" / "instates.nc")

    # Set it to read mode
    mock_model = mock_model_factory(path=model_copy_path, mode="r")

    # Setup the component
    component = WflowStatesComponent(mock_model)

    # Read the data and capture logs
    with caplog.at_level("WARNING"):
        component.read()

    # Check for warning about missing CRS
    assert "CRS not found in states data, setting to model CRS." in caplog.text
    assert component.data.raster.crs == mock_model.crs


def test_wflow_states_component_write(
    mock_model: WflowSbmModel,
    grid_dummy_data: xr.DataArray,
):
    # Setup the component
    component = WflowStatesComponent(mock_model)
    mock_model.components = {"states": component}

    component._data = grid_dummy_data.to_dataset(name="test_layer")

    # Write to a file
    type(component.model.config).get_value = MagicMock(return_value="")  # dir_input
    component.write()

    # Check if the file was created and has the expected content
    file = Path(component.root.path, component._filename)
    assert file.is_file()

    written_data = xr.open_dataset(file)
    assert "test_layer" in written_data
    # rename dims is True so x, y get renamed when writting
    assert "latitude" in written_data
    assert np.all(written_data["test_layer"].values == grid_dummy_data.values)


def test_wflow_states_component_equal(
    mock_model: WflowSbmModel,
    grid_dummy_data: xr.DataArray,
):
    # Setup the components
    component = WflowStatesComponent(mock_model)
    component2 = WflowStatesComponent(mock_model)

    # Update them manually
    component._data = grid_dummy_data.to_dataset()
    component2._data = grid_dummy_data.to_dataset()

    # Assert these are equal
    eq, errors = component.test_equal(component2)
    assert eq is True

    # Update component2 to make them not equal
    component2._data["test_layer"] = grid_dummy_data + 1  # Modify data

    # Assert unequal
    eq, errors = component.test_equal(component2)
    assert eq is False
    assert "Other grid has additional maps" in errors
    assert "test_layer" in errors["Other grid has additional maps"]



"""Test plugin model class against hydromt.models.model_api."""

import copy
from uuid import uuid4

from hydromt_wflow.wflow_sbm import WflowSbmModel


def inner_benchmark_loop(tmpdir, wflow_ini):
    # don't interfere with output of other rounds
    root = tmpdir.join(str(uuid4()))
    mod1 = WflowSbmModel(root=root, mode="w", data_libs="artifact_data")

    # Build model
    mod1.build(steps=copy.deepcopy(wflow_ini))


def test_outer_benchmark(tmpdir, wflow_ini, benchmark):
    benchmark.pedantic(
        inner_benchmark_loop, (tmpdir, wflow_ini), iterations=1, rounds=10
    )



import logging
from pathlib import Path
from typing import cast

import geopandas as gpd
from geopandas.testing import assert_geodataframe_equal
from hydromt.model import Model
from hydromt.model.components import GeomsComponent, ModelComponent
from hydromt.model.steps import hydromt_step

logger = logging.getLogger(f"hydromt.{__name__}")


class WflowGeomsComponent(GeomsComponent):
    """Wflow Geoms Component to manage spatial geometries.

    It extends the base GeomsComponent from hydromt and consists of a dictionary of
    geopandas GeoDataFrames.
    """

    def __init__(
        self,
        model: Model,
        *,
        filename: str = "staticgeoms/{name}.geojson",
        region_component: str | None = None,
        region_filename: str = "staticgeoms/geoms_region.geojson",
    ):
        """Initialize a WflowGeomsComponent.

        Parameters
        ----------
        model : Model
            HydroMT model instance
        filename : str
            The path to use for reading and writing of component data by default.
            by default "staticgeoms/{name}.geojson", i.e. one file per geodataframe in
            the data dictionary.
        region_component : str, optional
            The name of the region component to use as reference for this component's
            region. If None, the region will be set to the union of all geometries in
            the data dictionary.
        region_filename : str
            The path to use for writing the region data to a file. By default
            "staticgeoms/geoms_region.geojson".
        """
        super().__init__(
            model=model,
            filename=filename,
            region_component=region_component,
            region_filename=region_filename,
        )

    @hydromt_step
    def read(
        self,
        folder: str = "staticgeoms",
    ):
        """
        Read static geometries and adds to ``geoms``.

        If ``dir_input`` is set in the config, the path where all static geometries are
        read, will be constructed as ``<model_root>/<dir_input>/<geoms_fn>``.
        Where <dir_input> is relative to the model root. Depending on the config value
        ``dir_input``, the path will be constructed differently.

        Parameters
        ----------
        folder : str, optional
            Folder name/path where the static geometries are stored relative to the
            model root and ``dir_input`` if any. By default "staticgeoms".
        """
        # Check for input dir
        p_input = Path(self.model.config.get_value("dir_input", fallback=""), folder)
        pattern = Path(p_input, "{name}.geojson")

        super().read(filename=str(pattern))

    @hydromt_step
    def write(
        self,
        folder: str = "staticgeoms",
        to_wgs84: bool = False,
        precision: int | None = None,
        **kwargs,
    ) -> None:
        r"""
        Write model geometries to vector file(s) (by default GeoJSON) at <dir_out>/\*.geojson.

        Checks the path of ``geoms_fn`` using both model root and
        ``dir_input``. If not found uses the default path ``staticgeoms`` in the root
        folder.

        Key-word arguments are passed to :py:meth:`geopandas.GeoDataFrame.to_file`

        Parameters
        ----------
        folder : Path, optional
            The directory to write the geometry files to. If it does not exist, it will
            be created.
        to_wgs84 : bool, optional
            If True, the geoms will be reprojected to WGS84(EPSG:4326) before being
            written.
        precision : int, optional
            The precision to use for writing the geometries. If None, it will be set to 1
            for projected CRS and 6 for geographic CRS.
        **kwargs : dict
            Additional keyword arguments that are passed to the
            `geopandas.to_file` function.
        """  # noqa: E501
        # Check for dir_input
        p_input = Path(self.model.config.get_value("dir_input", fallback=""), folder)
        pattern = Path(p_input, "{name}.geojson")

        # Set precision
        if precision is None:
            if self.model.crs.is_projected:
                _precision = 1
            else:
                _precision = 6
        else:
            _precision = precision

        grid_size = 10 ** (-_precision)
        for gdf in self.data.values():
            gdf.geometry = gdf.geometry.set_precision(
                grid_size=grid_size,
            )

        super().write(
            filename=str(pattern),
            to_wgs84=to_wgs84,
            **kwargs,
        )

    def get(self, name: str) -> gpd.GeoDataFrame | gpd.GeoSeries:
        """Get geometry by name.

        Parameters
        ----------
        name : str
            The name of the geometry to retrieve.

        Returns
        -------
        gpd.GeoDataFrame | gpd.GeoSeries
            The geometry associated with the given name.

        Raises
        ------
        KeyError
            If the geometry with the specified name does not exist.
        """
        if name not in self.data:
            raise KeyError(
                f"Geometry '{name}' not found in geoms. Available geometries: {list(self.data.keys())}"  # noqa: E501
            )
        return self.data[name]

    def pop(self, name: str) -> gpd.GeoDataFrame | gpd.GeoSeries:
        """Remove and return geometry by name.

        Parameters
        ----------
        name : str
            The name of the geometry to remove and return.

        Returns
        -------
        gpd.GeoDataFrame | gpd.GeoSeries
            The geometry associated with the given name.

        Raises
        ------
        KeyError
            If the geometry with the specified name does not exist.
        """
        if name not in self.data:
            raise KeyError(
                f"Geometry '{name}' not found in geoms. Available geometries: {list(self.data.keys())}"  # noqa: E501
            )
        geom = self.data.pop(name)
        logger.debug(f"Removed geometry '{name}' from geoms.")
        return geom

    def clear(self) -> None:
        """Clear all geometries."""
        self.data.clear()
        logger.debug("Cleared all geometries from geoms.")

    def test_equal(self, other: ModelComponent) -> tuple[bool, dict[str, str]]:
        """Test if two GeomsComponents are equal.

        Parameters
        ----------
        other: GeomsComponent
            The other GeomsComponent to compare with.

        Returns
        -------
        tuple[bool, dict[str, str]]
            True if the components are equal, and a dict with the associated errors per
            property checked.
        """
        errors: dict[str, str] = {}
        if not isinstance(other, self.__class__):
            errors["__class__"] = f"other does not inherit from {self.__class__}."

        other_geoms = cast(WflowGeomsComponent, other)
        for name, gdf in self.data.items():
            if name not in other_geoms.data:
                errors[name] = "Geom not found in other component."

            # Geometries need to be normalized before comparison
            # (eg polygon points defined in the exact same order)
            gdf = gdf.copy()
            gdf["geometry"] = gdf["geometry"].apply(lambda geom: geom.normalize())
            other_gdf = other_geoms.data[name].copy()
            other_gdf["geometry"] = other_gdf["geometry"].apply(
                lambda geom: geom.normalize()
            )

            try:
                assert_geodataframe_equal(
                    gdf,
                    other_gdf,
                    check_like=True,
                    check_less_precise=True,
                )
            except AssertionError as e:
                errors[name] = str(e)

        return len(errors) == 0, errors



import geopandas as gpd
import numpy as np
import pytest
import rasterio
import xarray as xr
from shapely.geometry import box

from hydromt_wflow.workflows import glaciermaps


@pytest.fixture
def sample_gdf():
    """Return a simple GeoDataFrame with glacier polygons and AREA attribute."""
    geometries = [box(0, 0, 1, 1), box(1, 1, 2, 2)]
    data = {"simple_id": [1, 2], "AREA": [1.0, 2.0]}
    return gpd.GeoDataFrame(data, geometry=geometries, crs="EPSG:4326")


@pytest.fixture
def sample_dataset() -> xr.Dataset:
    """Return a dummy xr.Dataset with raster attributes and a fake raster accessor."""
    # Create a simple 10x10 grid
    data = np.zeros((10, 10), dtype=np.float32)
    transform = rasterio.transform.from_origin(0, 10, 1, 1)

    # Create an xarray DataArray with rioxarray accessor
    da = xr.DataArray(
        data,
        dims=("y", "x"),
        coords={"y": np.arange(10, 0, -1), "x": np.arange(10)},
    )
    da.rio.write_crs("EPSG:4326", inplace=True)
    da.rio.write_transform(transform, inplace=True)
    da.rio.write_nodata(-9999, inplace=True)

    return da.to_dataset(name="elevtn")


def test_glaciermaps_output_structure(sample_gdf, sample_dataset):
    """Test that glaciermaps returns expected variables and correct shapes."""
    ds_out = glaciermaps(
        gdf=sample_gdf,
        ds_like=sample_dataset,
        id_column="simple_id",
        elevtn_name="elevtn",
    )

    # Check output type
    assert isinstance(ds_out, xr.Dataset)

    # Check expected variables
    for var in ["glacareas", "glacstore", "glacfracs"]:
        assert var in ds_out

    # Check shapes
    expected_shape = sample_dataset["elevtn"].shape
    for var in ["glacareas", "glacstore", "glacfracs"]:
        assert ds_out[var].shape == expected_shape

    # Check data types
    assert ds_out["glacareas"].dtype == np.int32
    assert ds_out["glacstore"].dtype == np.float32
    assert ds_out["glacfracs"].dtype == np.float32


def test_glaciermaps_nodata_and_values(sample_gdf, sample_dataset):
    """Test that nodata values are set and maps are non-negative."""
    ds_out = glaciermaps(
        gdf=sample_gdf,
        ds_like=sample_dataset,
        id_column="simple_id",
        elevtn_name="elevtn",
    )

    # Nodata should be 0
    for var in ["glacareas", "glacstore", "glacfracs"]:
        nodata_value = ds_out[var].rio.nodata
        assert nodata_value == 0

    # Values should be >= 0
    for var in ["glacareas", "glacstore", "glacfracs"]:
        data = ds_out[var].values
        assert np.all(data >= 0)


def test_glaciermaps_respects_id_column(sample_gdf, sample_dataset):
    """Test that changing id_column works."""
    sample_gdf["custom_id"] = sample_gdf["simple_id"] + 100
    ds_out = glaciermaps(
        gdf=sample_gdf,
        ds_like=sample_dataset,
        id_column="custom_id",
        elevtn_name="elevtn",
    )
    assert "glacareas" in ds_out

    expected_ids = set(sample_gdf["custom_id"]) | {0}  # 0 for nodata
    actual_ids = set(np.unique(ds_out["glacareas"].values))
    assert actual_ids <= expected_ids, (
        "All elements of actual_ids should be in expected_ids"
    )


def test_glaciermaps_with_projected_crs(sample_gdf, sample_dataset):
    """Test glaciermaps using a projected CRS (EPSG:3857) in ds_like."""
    # Reproject ds_like to EPSG:3857 (Web Mercator)
    ds_proj = sample_dataset.rio.reproject("EPSG:3857")

    # Ensure sample_gdf is still in EPSG:4326 to test reprojection inside glaciermaps
    assert sample_gdf.crs.to_string() == "EPSG:4326"
    assert ds_proj.rio.crs.to_string() == "EPSG:3857"

    ds_out = glaciermaps(
        gdf=sample_gdf, ds_like=ds_proj, id_column="simple_id", elevtn_name="elevtn"
    )

    # Check CRS of output
    assert ds_out.rio.crs.to_string() == "EPSG:3857"

    # Check variable existence and types
    for var in ["glacareas", "glacstore", "glacfracs"]:
        assert var in ds_out
        assert ds_out[var].shape == ds_proj["elevtn"].shape

    # Check no NaNs or negative values
    assert np.all(ds_out["glacfracs"].values >= 0)
    assert np.all(ds_out["glacstore"].values >= 0)



from pathlib import Path

import numpy as np
import pytest
import xarray as xr

from hydromt_wflow.components.utils import (
    _mount,
    _relpath,
    get_mask_layer,
    make_config_paths_relative,
)
from hydromt_wflow.components.utils import test_equal_grid_data as equal_grid_data


def test__mount():
    # Call the function on unix path
    m = _mount("/d/tmp/foo")
    # Assert the mount
    assert m == "/d/"

    # Call the function on windows path
    m = _mount("d:/tmp/foo")
    # Assert the mount
    assert m == "d:/"

    # Call the function on a relative path
    m = _mount("tmp/foo")
    # Assert that it's None
    assert m is None


def test__relpath_abs(tmp_path: Path):
    # Call the function
    p = _relpath(Path(tmp_path, "tmp/tmp.txt"), tmp_path)

    # Assert the output
    assert isinstance(p, str)
    assert p == "tmp/tmp.txt"

    # Path one above the current, also pass as a string
    in_p = Path(tmp_path.parent, "tmp.txt").as_posix()
    p = _relpath(in_p, tmp_path)

    # Assert the output
    assert p == "../tmp.txt"


def test__relpath_rel(tmp_path: Path):
    # Call the function on a path that is already relative
    p = _relpath("tmp/tmp.txt", tmp_path)

    # Assert the output is just the same
    assert p == "tmp/tmp.txt"


def test__relpath_mount(tmp_path: Path, mount_string: str):
    # Call the function on a path that is located on another mount
    p = _relpath(Path(mount_string, "tmp", "tmp.txt"), tmp_path)

    # Assert the output is just the same
    assert p == f"{mount_string}tmp/tmp.txt"


def test__relpath_other(tmp_path: Path):
    # Call the function on value that could not be paths
    p = _relpath([2, 2], tmp_path)  # E.g. a list

    # Assert that the list is returned
    assert p == [2, 2]


def test_get_mask_layer_none(mask_layer: xr.DataArray):
    # Call the function, nothing in is nothing out
    mask = get_mask_layer(None)
    # Assert mask is indeed None
    assert mask is None

    # Call the function but layer not found in dataset(s)
    mask = get_mask_layer(
        "mask",
        mask_layer.to_dataset(name="no_mask"),
        mask_layer.to_dataset(name="still_no_mask"),
    )
    # Assert None because not found
    assert mask is None


def test_get_mask_layer_from_str(mask_layer: xr.DataArray):
    # Call the function with a layer in a dataset that can be found
    mask = get_mask_layer(
        "mask",
        mask_layer.to_dataset(name="mask"),
        mask_layer.to_dataset(name="no_mask"),
    )

    # Assert the output data
    assert isinstance(mask, xr.DataArray)
    np.testing.assert_array_equal(
        mask,
        np.array([[True, True], [True, False]]),
    )


def test_get_mask_layer_from_array(mask_layer: xr.DataArray):
    # Assert the current state of the mask layer
    np.testing.assert_array_equal(mask_layer.data, np.array([[1, 1], [1, -9999]]))
    assert mask_layer.raster.nodata == -9999

    # Call the function
    mask = get_mask_layer(mask_layer)

    # Assert its now boolean data
    np.testing.assert_array_equal(
        mask,
        np.array([[True, True], [True, False]]),
    )


def test_get_mask_layer_errors():
    # Call the function but give a faulty type, like a list
    with pytest.raises(
        ValueError,
        match="Unknown type for determining mask: list",
    ):
        _ = get_mask_layer([2, 2])


def test_make_config_paths_relative(
    tmp_path: Path,
    config_dummy_document: dict,
):
    # Assert that a full path is present
    p = config_dummy_document["baz"]["file1"]
    assert Path(p).is_absolute()
    assert config_dummy_document["spooky"]["ghost"] == [1, 2, 3]
    assert config_dummy_document["baz"]["file2"] == "tmp/tmp.txt"

    # Call the function
    cfg = make_config_paths_relative(config_dummy_document, tmp_path)

    # Assert the outcome
    # Assert that a full path is present
    p = cfg["baz"]["file1"]
    assert not Path(p).is_absolute()  # Not anymore
    assert cfg["spooky"]["ghost"] == [1, 2, 3]
    assert cfg["baz"]["file2"] == "tmp/tmp.txt"


def test__equal_grid():
    # Create two identical grids
    data1 = xr.DataArray(
        np.array([[1, 2], [3, 4]]),
        dims=("y", "x"),
        coords={
            "x": ("x", [0.5, 1.5]),
            "y": ("y", [0.5, 1.5]),
        },
    )
    data1.raster.set_crs(4326)
    data1.raster.set_nodata(-1)
    data1 = data1.to_dataset(name="data1")

    # Test equality
    eq, _ = equal_grid_data(data1, data1.copy())
    assert eq

    # Check for empty
    empty_grid = xr.Dataset()
    eq, errors = equal_grid_data(empty_grid, empty_grid)
    assert eq

    eq, errors = equal_grid_data(empty_grid, data1)
    assert not eq
    assert errors["grid"] == "first grid is empty, second is not"

    # Check for crs
    data2 = data1.copy()
    data2.raster.set_crs(3857)
    eq, errors = equal_grid_data(data1, data2)
    assert not eq
    assert errors["crs"] == "the two grids have different crs"

    # Check for dims
    data2 = data1.copy()
    data2 = data1.rename({"x": "x2"})
    eq, errors = equal_grid_data(data1, data2)
    assert not eq
    assert errors["dims"] == "dim x2 not in grid"

    data2 = data1.copy()
    data2["x"] = [0.5, 2.5]
    eq, errors = equal_grid_data(data1, data2)
    assert not eq
    assert errors["dims"] == "dim x not identical"

    # Check for missing data variables
    data2 = data1.copy()
    data2["data2"] = data2["data1"]
    eq, errors = equal_grid_data(data1, data2)
    assert not eq
    assert errors["Other grid has additional maps"] == "data2"

    eq, errors = equal_grid_data(data2, data1)
    assert not eq
    assert errors["Other grid is missing maps"] == "data2"

    # Check on data variable differences: dtype, nodata, values
    data2 = data1.copy()
    data2["data1"] = data2["data1"].astype(np.float32)
    eq, errors = equal_grid_data(data1, data2)
    assert not eq
    assert errors["1 invalid maps"] == {"data1": "float32 instead of int64"}

    data2 = data1.copy()
    data2["data1"].raster.set_nodata(-9999)
    eq, errors = equal_grid_data(data1, data2)
    assert not eq
    assert errors["1 invalid maps"] == {"data1": "nodata -9999 instead of -1; "}

    data2 = data1.copy()
    data2["data1"].values = [[1, 1], [3, 6]]
    eq, errors = equal_grid_data(data1, data2)
    assert not eq
    assert errors["1 invalid maps"] == {"data1": "mean diff (2 cells): -0.5000; "}
